Require Import Le Plus List Coq.Arith.Compare Coq.Program.Tactics.
Require Import EnvTy ILM ILM_Types.

Open Scope map_scope.

Set Implicit Arguments.

(* This file contains the proof of type preservation: Theorem 1 *)


(* Some of the lemmas to prove preservation of the memory-type related conditions
   do not actually need the alias tokens to be satisfied by E (some of them do not even talk about
   E at all), however a basic condition must be fulfilled: The tokens do not allow the drivation
   of "a != a" for any a.
   This is trivially true for any set of tokens satisfied by some E. *)
Definition alias_tokens_possible (G: alias_tokens):Prop
 := forall a, ~alias G (AliasNeq a a).

(* Some useful facts about possible alias tokns *)
Lemma alias_tokens_possible_eq_neq G a1 a2:
alias_tokens_possible G -> alias G (AliasEq a1 a2) -> ~alias G (AliasNeq a1 a2).
Proof.
intros poss EQ NEQ. apply (poss a2). eapply AliasEqNeq.
- eapply AliasEqSym. eassumption.
- eassumption.
Qed.

Lemma alias_tokens_possible_incl G1 G2: G1 ⊆ G2 -> alias_tokens_possible G2 -> alias_tokens_possible G1.
Proof.
intros SUBS POSS a ALIAS. eapply (alias_incl SUBS) in ALIAS. apply (POSS a). assumption.
Qed.

Lemma alias_disjoint_disjoint G (poss:alias_tokens_possible G) A1 A2:
  alias_disjoint G A1 A2 -> disjoint A1 A2.
Proof.
intros adisj x [IN1 IN2]. unfold alias_disjoint, alias_tokens_possible in *.
apply (poss x). crush.
Qed.

Lemma alias_disjoint_incl G1 G2: G1 ⊆ G2 -> forall A1 A2, alias_disjoint G1 A1 A2 -> alias_disjoint G2 A1 A2.
Proof.
intros. intros a1 a2 IN1 IN2. eapply alias_incl; crush.
Qed.

Lemma alias_tokens_apply_possible E G:
  alias_tokens_apply E G -> alias_tokens_possible G.
Proof.
intros appl x do_alias.
assert (alias_token_applies E (AliasNeq x x)) by eauto using alias_sound.
simpl in H. apply H. reflexivity.
Qed.
Hint Resolve alias_tokens_apply_possible.




(** Memory type validity *)
(* type validity is preserved by adding tokens *)
Lemma memMoreTokensValid ET ST G G' F: G ⊆ G' ->
  memTypesValid ET ST G F -> memTypesValid ET ST G' F.
Proof.
intros subs valid. invc valid. apply MTV; try assumption.
- intros. eapply alias_disjoint_incl; crush.
Qed.

(* type validity is preserved by focusing on a new set *)
Lemma memFocusValid (ET:onv ty) (ST:onv memory_type) G F m A
  (ptr:forall a, a ∈ A -> exists t, ET[a] = Some (ptrTo t))
  (new:ST[m] = None)  (disj:forall m' A', ST[m'] = Some (MemFocus A') -> alias_disjoint G A' A):
  memTypesValid ET ST G F -> memTypesValid ET (ST[m <- Some (MemFocus A)]) G (F ∪ A).
Proof.
intros valid. invc valid. apply MTV; try assumption.
- crush.
- intros a IN. cset_tac. destruct IN as [IN|IN].
  + now eauto using focus_sane.
  + now eauto using ptr.
- destruct unique_mtop as [mtop [ex un]]. exists mtop. split.
  + destruct_lookup_a. crush.
  + intros. destruct_lookup_a. crush.
- intros. destruct_lookup_a. destruct_lookup.
  + autoinjections. eapply alias_disjoint_sym. crush.
  + crush.
- intros. destruct_lookup_a. now eauto.
- intros a IN. cset_tac. destruct IN as [IN|IN].
  + destruct (focus_in_mem a IN) as [m' [A' [EQ' IN']]]. exists m'. exists A'. destruct_lookup_a. crush.
  + exists m. exists A. crush.
Qed.
Lemma memReFocusValid (ET:onv ty) (ST:onv memory_type) G F m A
  (inF: A ⊆ F) (ptr:forall a, a ∈ A -> exists t, ET[a] = Some (ptrTo t))
  (new:ST[m] = None)  (disj:forall m' A', ST[m'] = Some (MemFocus A') -> alias_disjoint G A' A):
  memTypesValid ET ST G (F\A) -> memTypesValid ET (ST[m <- Some (MemFocus A)]) G F.
Proof.
assert (H: F = (F\A) ∪ A). { rewrite union_comm. apply incl_set_decomp. assumption. }
rewrite H at 2. apply memFocusValid; assumption.
Qed.

(* type validity is preserved by unfocusing a memory *)
Lemma memUnfocusValid (ET:onv ty) (ST:onv memory_type) G F m A
  (poss:alias_tokens_possible G) (isFocus:ST[m] = Some (MemFocus A)):
  memTypesValid ET ST G F -> memTypesValid ET (ST[m <- None]) G (F \ A).
Proof.
intros valid. invc valid. apply MTV; try assumption.
- crush.
- intros a IN. cset_tac. now eauto using focus_sane.
- destruct unique_mtop as [mtop [ex un]]. exists mtop. split; crush.
- intros. destruct_lookup_a. destruct_lookup_a. now eauto.
- intros. destruct_lookup_a. assert (DISJ:disjoint A0 A) by eauto using alias_disjoint_disjoint.
  intros x. cset_tac. eapply mems_in_focus; now eauto. unfold disjoint in *. eapply DISJ. now eauto.
- intros a IN. assert (a ∈ F) by (eapply minus_incl; crush). destruct (focus_in_mem _ H) as [m' [A' [EQ' IN']]].
  exists m'. exists A'. split.
  + destruct_lookup.
    * exfalso. rewrite EQ' in *. autoinjections. eapply minus_in_in; crush.
    * crush.
  + crush.
Qed.

(* type validity is preserved by renaming the panorama memory *)
Lemma memRenamePanoramaValid ET (ST:onv memory_type) G F m1 m2 (wasmtop:ST[m1] = Some MemPanorama) (wasfree:ST[m1 <- None][m2] = None):
  memTypesValid ET ST G F -> memTypesValid ET (ST[m1 <- None][m2 <- Some MemPanorama]) G F.
Proof.
intros valid. invc valid. apply MTV; try assumption.
- crush.
- exists m2. split.
  + simplify lookup. reflexivity.
  + intros x EQ. destruct_lookup_a. destruct_lookup_a.
    destruct unique_mtop as [mtop [ex un]]. apply un in EQ. apply un in wasmtop. subst. now isabsurd.
- intros. now repeat (destruct_lookup_a; eauto using focus_disjoint).
- intros m a EQ. destruct_lookup_a. destruct_lookup_a. eapply mems_in_focus; now eauto.
- intros a IN. destruct (focus_in_mem _ IN) as [m' [A' [EQ' IN']]]. exists m'. exists A'. crush.
Qed.

(* type validity is preserved by renaming any memory *)
Lemma memRenameValid ET (ST:onv memory_type) G F m1 m2 t (wastype:ST[m1] = Some t)
  (wasfree:ST[m1 <- None][m2] = None):
  memTypesValid ET ST G F -> memTypesValid ET (ST[m1 <- None][m2 <- Some t]) G F.
Proof.
destruct t.
- intros valid. invc valid. apply MTV; try assumption.
  + crush.
  + destruct unique_mtop as [mtop [ex un]]. exists mtop. split.
    * destruct_lookup_a. crush.
    * intros x EQ. destruct_lookup_a. destruct_lookup_a. now eapply un.
  + intros. destruct_lookup.
    * destruct_lookup_a. eapply (focus_disjoint m0 _ m1); rewrite ?wastype; now eauto.
    * destruct_lookup_a. destruct_lookup_b. eapply (focus_disjoint m1 _ m3); rewrite ?wastype; now eauto.
  + intros m A EQ. destruct_lookup.
    * eapply (mems_in_focus m1). rewrite wastype. now eauto.
    * destruct_lookup_a. now eauto.
  + intros a IN. destruct (focus_in_mem _ IN) as [m' [A' [EQ' IN']]]. destruct_prop(m'=m1). 
    * exists m2. exists A'. simplify lookup. split; crush.
    * exists m'. exists A'. crush.
- now eauto using memRenamePanoramaValid.
Qed.

(* type validity is preserved by adding a variable *)
Lemma memUpdateValid (ET:onv ty) (ST:onv memory_type) G F x t (new:ET[x] = None):
  memTypesValid ET ST G F -> memTypesValid (ET[x <- Some t]) ST G F.
Proof.
intros valid. invc valid. apply MTV; try assumption.
- intros a IN. destruct (focus_sane a IN) as [pt A]. exists pt. destruct_lookup_a. assumption.
Qed.

(* type validity is preserved by merging two focused memories *)
Lemma memMergeFocusValid (ET:onv ty) (ST:onv memory_type) G F m1 A1 m2 A2 m3 (poss:alias_tokens_possible G)
  (ineq:m1 <> m2) (ism1:ST[m1] = Some (MemFocus A1)) (ism2:ST[m2] = Some (MemFocus A2))
  (new:ST[m1 <- None][m2 <- None][m3] = None):
  memTypesValid ET ST G F -> memTypesValid ET (ST[m1 <- None][m2 <- None][m3 <- Some (MemFocus (A1 ∪ A2))]) G F.
Proof.
intros valid. inversion valid. eapply memReFocusValid.
- crush.
- intros a IN. cset_tac. destruct IN.
  + eapply focus_sane, (mems_in_focus m1); crush.
  + eapply focus_sane, (mems_in_focus m2); crush.
- crush.
- intros. destruct_lookup_a. destruct_lookup_a.
  unfold alias_disjoint in *. intros. cset_tac. destruct H1.
  + eapply (focus_disjoint m' _ m1); crush.
  + eapply (focus_disjoint m' _ m2); crush.
- rewrite <- minus_union. eapply memUnfocusValid, memUnfocusValid; crush.
Qed.

(* type validity is preserved by splitting a focussed memory appropriately *)
Lemma memSplitFocusValid ET (ST:onv memory_type) G F m1 A1 A m2 m3 (ineq:m2 <> m3)
  (ism1:ST[m1] = Some (MemFocus A1)) (new2: ST[m1 <- None][m2] = None) (new3: ST[m1 <- None][m3] = None)
  (inA1: A ⊆ A1) (disj:alias_disjoint G A (A1\A)) (poss:alias_tokens_possible G):
  memTypesValid ET ST G F -> memTypesValid ET (ST[m1 <- None][m2 <- Some (MemFocus (A1\A))][m3 <- Some (MemFocus A)]) G F.
Proof.
intros valid. inversion valid. eapply memReFocusValid.
- eapply incl_trans; crush.
- intros a IN. apply focus_sane. eapply incl_trans; crush.
- crush.
- intros. destruct_lookup.
  + autoinjections. apply alias_disjoint_sym. assumption.
  + destruct_lookup_a. eapply alias_disjoint_incl_r; crush.
- eapply memReFocusValid.
  + intros x IN. rewrite minus_spec. bool_to_prop. split.
    * apply incl_minus in IN. eapply mems_in_focus; crush.
    * cset_tac. isabsurd.
  + intros a IN. apply focus_sane. apply incl_minus in IN. eapply mems_in_focus; crush.
  + crush.
  + intros. destruct_lookup_a. eapply alias_disjoint_incl_r; crush.
  + assert (H: ((F\A) \ (A1\A)) = F \ A1). { rewrite minus_union. f_equal. symmetry. apply incl_set_decomp. crush. }
    rewrite H. eapply memUnfocusValid; crush.
Qed.


(** well-typed memories *)
(* typing relation is preserved by adding valid tokens *)
Lemma memMoreTokensTypes (ET:onv ty) (ST:onv memory_type) G G' F E S U: G ⊆ G' ->
  memoriesOfType ET ST G F E S U -> memoriesOfType ET ST G' F E S U.
Proof.
intros subs types. invc types. apply MOT; try assumption.
- now eauto using memMoreTokensValid.
Qed.

(* typing relation is preserved by adding a new variable *)
Lemma memUpdateTypes (ET:onv ty) (ST:onv memory_type) G F E S U x t v (new:ET[x] = None):
  memoriesOfType ET ST G F E S U -> memoriesOfType (ET[x <- Some t]) ST G F (E[x <- v]) S U.
Proof.
intros types. invc types. inv types_valid. apply MOT; try assumption.
- now eauto using memUpdateValid.
- intros. specialize (mems_type m mt H). destruct mt; simpl in *.
  + specialize (mems_in_focus _ _ H). rewrite lookup_set_update.
    * crush.
    * intros IN. apply mems_in_focus in IN. destruct (focus_sane _ IN) as [t' EQ]. congruence.
  + rewrite lookup_set_update.
    * crush.
    * intros IN. destruct (focus_sane _ IN). congruence.
Qed.

(* typing relation is preserved by adding a new memory disjoint to the old ones *)
Lemma memNewMemTypes (ET:onv ty) (ST:onv memory_type) G F E S U m A (mv:mem) (tnew:ST[m] = None)
  (tval:forall a, a ∈ A -> exists t, ET[a] = Some (ptrTo t)) (disj:alias_disjoint G A F)
  (vval:domain mv ⊆ lookup_set E A) (vnew:disjoint (lookup_set E A) U) (def:default mv = C_Unalloc):
  memoriesOfType ET ST G F E S U -> memoriesOfType ET (ST[m <- Some (MemFocus A)]) G (F ∪ A) E (S[m <- mv]) (U ∪ (domain mv)).
Proof.
intros types. invc types. apply MOT.
- eapply memFocusValid; now eauto using disjointF_disjointAll.
- crush.
- intros. destruct_lookup_a. now eauto using incl_trans, mems_in_U.
- intros. destruct_lookup_a. specialize (mems_type m0 mt H). destruct mt; simpl in *.
  + assumption.
  + rewrite lookup_set_union. apply disjoint_split.
    * assumption.
    * apply disjoint_sym. eauto using disjoint_incl_r, mems_in_U.
Qed.

(* special case of memNewMemTypes for adding a singleton memory (containing just one element) *)
Lemma memNewSingletonMemTypes (ET:onv ty) (ST:onv memory_type) G F E S U m a t v (tval:ET[a] = Some (ptrTo t))
  (tnew:ST[m] = None) (disj:alias_disjoint G {{a}} F) (isval: E[a] = v) (vnew: v ∉ U):
  memoriesOfType ET ST G F E S U -> memoriesOfType ET (ST[m <- Some (MemFocus {{a}})]) G (F ∪ {{a}}) E (S[m <- (mempty[v <- C_Uninit])]) (U ∪ {{v}}).
Proof.
intros types. unfold mempty.
assert (H:{{v}} = domain (mempty[v <- C_Uninit])).
{ unfold mempty. rewrite domain_update by (rewrite default_spec; crush). rewrite domain_empty. cset_tac; tauto. }
rewrite H. eapply memNewMemTypes; unfold mempty; try eassumption.
- intros b IN. apply element_single in IN. exists t. crush.
- rewrite domain_update by (rewrite default_spec; crush). rewrite domain_empty, empty_neutral_union.
  intros x IN. apply element_single in IN. apply lookup_set_spec. exists a. split.
  + cset_tac; left; reflexivity.
  + crush.
- intros x [IN1 IN2]. apply lookup_set_spec in IN1. destruct IN1 as [y [IN1 LOOKUP]]. apply element_single in IN1.
  crush.
- rewrite default_spec2. rewrite default_spec. reflexivity.
Qed.


(* typing relation is preserved when merging two memories *)
Lemma memMergeFocusTypes ET (ST:onv memory_type) G F E S U m1 A1 m2 A2 m3 (ineq:m1 <> m2)
  (ism1:ST[m1] = Some (MemFocus A1)) (ism2:ST[m2] = Some (MemFocus A2))
  (new:ST[m1 <- None][m2 <- None][m3] = None) (tokens_possible: alias_tokens_possible G):
  memoriesOfType ET ST G F E S U -> memoriesOfType ET (ST[m1 <- None][m2 <- None][m3 <- Some (MemFocus (A1 ∪ A2))]) G F E (S[m3 <- merge_mem (S[m1]) (S[m2])]) U.
Proof.
intros types. inv types. apply MOT.
- eapply memMergeFocusValid; crush.
- intros. destruct_lookup_b. rewrite merge_mem_default. crush.
- intros. destruct_lookup_b. rewrite merge_mem_domain by eauto. crush.
- intros. destruct_lookup.
  + (* well-typedness of new memory *) autoinjections. eapply mergedFocusedMemTypes; crush.
  + (* well-typedness of old memory *) destruct_lookup_a. destruct_lookup_a. crush.
Qed.
Lemma memMergePanoramaTypes ET (ST:onv memory_type) G F E S U m1 A1 m2 m3
  (ism1:ST[m1] = Some (MemFocus A1)) (ism2:ST[m2] = Some MemPanorama)
  (new:ST[m1 <- None][m2 <- None][m3] = None) (tokens_apply: alias_tokens_apply E G):
  memoriesOfType ET ST G F E S U -> memoriesOfType ET (ST[m1 <- None][m2 <- None][m3 <- Some MemPanorama]) G (F\A1) E (S[m3 <- merge_mem (S[m1]) (S[m2])]) U.
Proof.
intros types. inv types. apply MOT.
- eapply memRenamePanoramaValid.
  + crush.
  + crush.
  + eapply memUnfocusValid; eauto.
- intros. destruct_lookup_b. rewrite merge_mem_default. eauto.
- intros. destruct_lookup_b. rewrite merge_mem_domain by eauto. crush.
- intros. destruct_lookup.
  + (* well-typedness of new memory *) autoinjections. eapply mergedPanoramaTypes; eauto using focus_disjoint_rest.
  + (* well-typedness of old memory *) destruct_lookup_a. destruct_lookup_a. destruct mt; simpl in *.
    * specialize (mems_type _ _ H). eauto.
    * exfalso. apply n0. eapply mtop_unique; eauto.
Qed.


(* Typing relation is preserved when splitting two memories *)
Lemma memSplitFocusTypes ET (ST:onv memory_type) G F E S U m1 A1 A m2 m3 (ineq:m2 <> m3) (inA1: A ⊆ A1)
  (ism1:ST[m1] = Some (MemFocus A1)) (new2: ST[m1 <- None][m2] = None) (new3: ST[m1 <- None][m3] = None)
  (disj:alias_disjoint G A (A1\A)) (tokens_possible: alias_tokens_possible G):
  memoriesOfType ET ST G F E S U -> memoriesOfType ET (ST[m1 <- None][m2 <- Some (MemFocus (A1\A))][m3 <- Some (MemFocus A)]) G F E (S[m2 <- restrict_mem (S[m1]) (domain (S[m1]) \ lookup_set E A)][m3 <- restrict_mem (S[m1]) (lookup_set E A)]) U.
Proof.
intros types. inv types. split.
- eapply memSplitFocusValid; crush.
- intros. destruct_lookup; [|destruct_lookup_b]; rewrite restrict_mem_default; reflexivity.
- intros. destruct_lookup; [|destruct_lookup_b]; rewrite restrict_mem_domain by crush; eapply meet_incl; crush.
- intros. destruct_lookup; [|destruct_lookup].
  + (* well-typedness of m3 *) autoinjections. eapply splitMemoryTypes; eauto.
  + (* well-typedness of m2 *) autoinjections. eapply splitFocusRemainderTypes; eauto.
  + (* well-typedness of old memory *) destruct_lookup_a. crush.
Qed.
Lemma memSplitPanoramaTypes ET (ST:onv memory_type) G F E S U m1 (A:cset var) m2 m3 (ineq:m2 <> m3)
  (ism1: ST[m1] = Some MemPanorama) (new2: ST[m1 <- None][m2] = None) (new3: ST[m1 <- None][m3] = None)
  (isptr:forall a, a ∈ A -> exists t, ET[a] = Some (ptrTo t))
  (disj:alias_disjoint G A F) (tokens_apply: alias_tokens_apply E G):
  memoriesOfType ET ST G F E S U -> memoriesOfType ET (ST[m1 <- None][m2 <- Some MemPanorama][m3 <- Some (MemFocus A)]) G (F ∪ A) E (S[m2 <- restrict_mem (S[m1]) (domain (S[m1]) \ lookup_set E A)][m3 <- restrict_mem (S[m1]) (lookup_set E A)]) U.
Proof.
intros types. inv types. split.
- eapply memFocusValid; try eassumption.
  + crush.
  + intros. destruct_lookup_a. destruct_lookup_a. eapply disjointF_disjointAll; eauto.
  + eapply memRenamePanoramaValid; crush.
- intros. destruct_lookup; [|destruct_lookup_b]; rewrite restrict_mem_default; reflexivity.
- intros. destruct_lookup; [|destruct_lookup_b]; rewrite restrict_mem_domain by eauto; eapply meet_incl; eauto.
- intros. destruct_lookup; [|destruct_lookup].
  + (* well-typedness of m3 *) autoinjections. eapply splitMemoryTypes; crush.
  + (* well-typedness of m2 *) autoinjections. eapply splitPanoramaRemainderTypes; crush.
  + (* well-typedness of old memory *) destruct_lookup_a. destruct mt; simpl in *.
    * specialize (mems_type _ _ H). crush.
    * exfalso. apply n1. eapply mtop_unique; eauto.
Qed.


(* Replacing a memory by another one with the same or smaller domain preserves the typing *)
Lemma memRenameTypes ET (ST:onv memory_type) G F E (S:env mem) U m1 m2 mt mv
  (ism1: ST[m1] = Some mt) (new: ST[m1 <- None][m2] = None)
  (dom: domain mv ⊆ domain (S[m1])) (samedef: default (S[m1]) = default mv):
  memoriesOfType ET ST G F E S U -> memoriesOfType ET (ST[m1 <- None][m2 <- Some mt]) G F E (S[m2 <- mv]) U.
Proof.
intros types. inv types. apply MOT.
- eapply memRenameValid; crush.
- intros. destruct_lookup_b. rewrite <- samedef. eauto.
- intros. destruct_lookup_b. eapply incl_trans; eauto.
- intros. destruct_lookup_b. autoinjections. specialize (mems_type _ _ ism1). destruct mt0; simpl in *.
  + eapply incl_trans; eauto.
  + eapply disjoint_incl_lr, mems_type; eauto using incl_refl.
Qed.


(** Memory oracle *)
(* Oracle properties are preserved by changing a non-pointer variable *)
Lemma oracleUpdateNoPtr (ET:onv ty) LT s E L U mv x t v (noptr:~isPtr t):
  oracleIsRight ET LT s E L U mv -> oracleIsRight (ET[x <- Some t]) LT s (E[x <- v]) L U mv.
Proof.
intros right. invc right. apply OIR; try assumption.
- intros. destruct_lookup_b. autoinjections. exfalso. apply noptr. now apply ptrIsPtr.
Qed.

(* Oracle properties are preserved by updating a pointer variable, if the new value
   has the proper type in the oracle *)
Lemma oracleUpdatePtr ET LT (s:memory_oracle) E L U mv a t v
  (iso:s[v] = Some t):
  oracleIsRight ET LT s E L U mv -> oracleIsRight (ET[a <- Some (ptrTo t)]) LT s (E[a <- v]) L U mv.
Proof.
intros right. inv right. apply OIR; try assumption.
- intros. destruct_lookup.
  + autoinjections. apply ptrTo_injective in H1. crush.
  + eapply pointers_e_valid; eauto.
Qed.

(* Oracle properties are preserved if the merged-all memory does not change *)
Lemma oracleSameMem ET LT s E L U mv1 mv2: mv1 === mv2 ->
  oracleIsRight ET LT s E L U mv1 -> oracleIsRight ET LT s E L U mv2.
Proof.
intros EQ right. invc right. split; try assumption.
- intros. specialize (oracle_types v t H). rewrite EQ in oracle_types. assumption.
- intros. rewrite <-EQ in H0. eauto.
Qed.

(* Oracle properties are preserved by adding a new location *)
Lemma oracleNewLocation ET LT (s:memory_oracle) E L U (mv1 mv2: mem) v t
  (upd: mv2 === mv1[v <- C_Uninit]) (new: v ∉ U):
  oracleIsRight ET LT s E L U mv1 -> oracleIsRight ET LT (s[v <- Some t]) E L (U ∪ {{v}}) mv2.
Proof.
intros right. invc right. split.
- intros. destruct_prop (v=v0); rewrite upd; simplify lookup'.
  + split.
    * cset_tac. right. left. reflexivity.
    * intros. discriminate.
  + destruct (oracle_types _ _ H) as [INU HASVAL]. split.
    * cset_tac. left. assumption.
    * assumption.
- intros ? ? H. specialize (pointers_e_valid _ _ H). destruct_lookup.
  + exfalso. specialize (oracle_types _ _ pointers_e_valid). apply new. tauto.
  + crush.
- intros ? ? ? ? ? GET1 GET2 H. specialize (pointers_lab_valid _ _ _ _ _ GET1 GET2 H). destruct_lookup.
  + (* new location in some label closure *) exfalso. specialize (oracle_types _ _ pointers_lab_valid).
    apply new. tauto.
  + crush.
- intros. rewrite upd in H0. destruct_prop (v=v0); simplify lookup'.
  + (* the new location is a pointer somewhere *) exfalso. crush.
  + destruct_prop(v=val); simplify lookup'.
    * (* an old location points to the new one *) exfalso. specialize (pointers_mem_valid _ _ H _ H0).
      specialize (oracle_types _ _ pointers_mem_valid). apply new. tauto.
    * (* an old location points to an old one *) crush.
Qed.

(* Oracle properties are preserved by updating a location given by a pointer with a value
   of the corresponding type *)
Lemma oracleUpdateMem (ET:onv ty) LT s (E:env val) L U mv1 mv2 (a:var) (v:val) x t
  (upd: mv2 === mv1[v <- C(E[x])])
  (aval:E[a] = v) (ahasty:ET[a] = Some (ptrTo t)) (xhasty: ET[x] = Some t) (valty: valOfType (E[x]) t):
  oracleIsRight ET LT s E L U mv1 -> oracleIsRight ET LT s E L U mv2.
Proof.
intros right. invc right. apply OIR; try assumption.
- intros. specialize (oracle_types _ _ H). rewrite upd. clear upd mv2. destruct_lookup.
  + split; [tauto|]. intros. specialize (pointers_e_valid _ _ ahasty). rewrite pointers_e_valid in *.
    autoinjections. assumption.
  + assumption.
- intros. rewrite upd in H0. clear upd mv2. destruct_lookup.
  + (* the changed location *) autoinjections. eapply pointers_e_valid. specialize (pointers_e_valid _ _ ahasty).
    crush.
  + eauto.
Qed.

(* Oracle properties are preserved by freeing a location *)
Lemma oracleFreeInMem (ET:onv ty) LT s (E:env val) L U mv1 mv2 (v:val)
  (upd: mv2 === mv1[v <- C_Unalloc]):
  oracleIsRight ET LT s E L U mv1 -> oracleIsRight ET LT s E L U mv2.
Proof.
intros right. invc right. apply OIR; try assumption.
- intros. specialize (oracle_types _ _ H). rewrite upd. clear upd mv2. destruct_lookup.
  + split; [tauto|]. intros. discriminate.
  + assumption.
- intros. rewrite upd in H0. clear upd mv2. destruct_lookup.
  + (* the changed location *) discriminate.
  + eauto.
Qed.

(* Oracle properties are preserved by adding a new closure (block) with a sub-environment of
   the current value environment *)
Lemma oracleAddLab ET LT s (E:env val) L U mv ET' Z ZT m t rt
  (isSub: subEnv ET' ET):
  oracleIsRight ET LT s E L U mv -> oracleIsRight ET ((TB ET' ZT rt)::LT) s E ((blockI E Z m t)::L) U mv.
Proof.
intros right. invc right. apply OIR; try assumption.
- intros ? ? ? ? ? GET1 GET2. destruct n; invc GET1; invc GET2; simpl.
  + intros HASTY. eapply pointers_e_valid, isSub. assumption.
  + eapply pointers_lab_valid; crush.
Qed.

(* Oracle properties are preserved by dropping some closures *)
Lemma oracleDropLab ET LT s E L U mv n:
  oracleIsRight ET LT s E L U mv -> oracleIsRight ET (drop n LT) s E (drop n L) U mv.
Proof.
intros right. invc right. split; try assumption.
- intros ? ? ? ? ? GET1 GET2. apply get_drop in GET1. apply get_drop in GET2. crush.
Qed.

(* Oracle properties are preserved by switching to the value environment of a closure
   and updating the values and types of the function arguments *)
Lemma oracleChangeEnv ET LT s E L U mv n E' ET' Z ZT m t rt Y
  (GET1: get LT n (TB ET' ZT rt)) (GET2: get L n (blockI E' Z m t)) (argsT: argsOfType ET Y ZT):
  oracleIsRight ET LT s E L U mv -> oracleIsRight (updET ET' ZT Z) LT s (updE E' E Z Y) L U mv.
Proof.
intros right. assert (isR:oracleIsRight ET' LT s E' L U mv).
{ invc right. split; try assumption.
  - intros. specialize (pointers_lab_valid a t0 _ _ _ GET2 GET1). simpl in *.
    apply pointers_lab_valid. assumption. }
clear- argsT right isR. revert Y ZT argsT. unfold updET, updE. induction Z; simpl; intros.
- assumption.
- invc argsT; simpl.
  + assumption.
  + invc right. apply OIR; try assumption. intros. destruct_lookup.
    * crush.
    * edestruct IHZ; crush.
Qed.

(* Oracle properties are preserved by changing the closures, as long as the value environments do not change *)
Lemma oracleChangeLabs ET LT s E L1 L2 U mv
  (sameEnvs: forall n blk2 tb, get L2 n blk2 -> get LT n tb ->
                               exists blk1, get L1 n blk1 /\ block_E blk1 = block_E blk2):
  oracleIsRight ET LT s E L1 U mv -> oracleIsRight ET LT s E L2 U mv.
Proof.
intros right. invc right. split; try assumption.
- intros ? ? ? ? ? GETL2 GETLT. destruct (sameEnvs _ _ _ GETL2 GETLT) as [blk1 [GETL1 EQE]].
  rewrite <-EQE. eapply pointers_lab_valid; eassumption.
Qed.


(** Type preservaton *)

(* One specific lemma for each statement *)
Lemma expPreservation ET LT ST F s E L S U G x e b rt
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtExp G x e b) rt)
  v (evals:exp_eval E e = Some v):
  exists et, stateConsistent (ET[x <- Some et], LT, ST, F) s (E[x <- v], L, S, U, b) rt.
Proof.
invc types. invc stmt_types.
exists et. split; try eassumption.
- eapply envOfType_update.
  + assumption.
  + edestruct exp_type_soundness as [v' [A B]]; try eassumption.
    * assert (v=v') by congruence; crush.
- eapply memMoreTokensTypes, memUpdateTypes; crush.
- eapply oracleUpdateNoPtr; crush.
Qed.

Lemma ifPreservation ET LT ST F s E L S U G x b1 b2 rt
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtIf G x b1 b2) rt):
   stateConsistent (ET, LT, ST, F) s (E, L, S, U, b1) rt /\
   stateConsistent (ET, LT, ST, F) s (E, L, S, U, b2) rt.
Proof.
invc types. invc stmt_types.
split; apply SOT; try assumption; eapply memMoreTokensTypes; crush.
Qed.

Lemma allocPreservation ET LT ST G F s E L S U a m b rt
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtAlloc G a m b) rt)
  (tokens_apply:alias_tokens_apply E G)
  (v:val) `(v = allocator U) (mv:mem) `(mv = mempty[v<-C_Uninit]):
  exists et, stateConsistent (ET[a <- Some (ptrTo et)], LT, ST[m <- Some (MemFocus {{a}})], F ∪ {{a}})
                       (s[v <- Some et]) (E[a <- v], L, S[m <- mv], U ∪ {{v}}, b) rt.
Proof.
invc types. invc stmt_types.
exists et. apply SOT; try assumption.
- now eauto using envOfType_update, allocatorGivesPtr.
- eapply memNewSingletonMemTypes; try eassumption.
  + crush.
  + intros a1 a2 IN1 IN2. apply element_single in IN1. crush.
  + crush.
  + now apply allocatorGivesNew.
  + eapply memMoreTokensTypes, memUpdateTypes; crush.
- eapply oracleUpdatePtr; try eassumption.
  + crush.
  + eapply oracleNewLocation with (mv1 := merge_all_mems ST S); try eassumption.
    * symmetry. eapply merge_all_mems_new_mem; eauto using allocInNoMem.
    * apply allocatorGivesNew.
Qed.

Lemma mergePreservation ET LT ST F s E L S U G m'' m m' b rt
    (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtMerge G m'' m m' b) rt)
    (tokens_apply:alias_tokens_apply E G)
    mv `(mv = merge_mem (S[m]) (S[m'])):
    exists mt A, stateConsistent (ET, LT, ST[m <- None][m' <- None][m'' <- Some mt], F\A) s (E, L, S[m'' <- mv], U, b) rt.
Proof.
invc types.
assert (m <> m'). { invc stmt_types; [assumption|]. intros EQ. crush. }
assert (disjoint (domain (S[m])) (domain (S[m']))).
{ invc stmt_types; eapply welltypedMemoriesAreDisjoint; crush. }
invc stmt_types.
- exists (MemFocus (mA ∪ mA')). exists BSet.empty. rewrite minus_empty. apply SOT; try assumption.
  + eapply memMergeFocusTypes; crush.
  + eapply oracleSameMem; [|eassumption]. eapply merge_all_mems_merge; crush.
- exists MemPanorama. exists mA. apply SOT; try assumption.
  + eapply memMergePanoramaTypes; crush. (* this requires the tokens to apply *)
  + eapply oracleSameMem; [|eassumption]. eapply merge_all_mems_merge; crush.
Qed.

Lemma loadPreservation ET LT ST F s E L S U G x m a b rt
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtLoad G x m a b) rt)
  (tokens_apply:alias_tokens_apply E G)
  v `(S[m][E[a]] = C v):
  exists et, stateConsistent (ET[x <- Some et], LT, ST, F) s (E[x <- v], L, S, U, b) rt.
Proof.
invc types. invc stmt_types. exists et. apply SOT; try assumption.
- eapply envOfType_update.
  + assumption.
  + eapply oracleLoadTypes; eauto.
- eapply memMoreTokensTypes, memUpdateTypes; crush.
- destruct_prop(isPtr et).
  + unfold isPtr in i. destruct (isPtr_sigma et) as [[pt EQ]|NEQ]; [|now isabsurd].
    subst et. eapply oracleUpdatePtr.
    * invc oracle_right. specialize (pointers_e_valid _ _ H12). eapply (pointers_mem_valid _ _ pointers_e_valid).
      symmetry. rewrite merge_all_mems_hasval by crush. eauto.
    * assumption.
  + eapply oracleUpdateNoPtr; crush.
Qed.

Lemma storePreservation ET LT ST F s E L S U G m' m a x b rt
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtStore G m' m a x b) rt)
  (tokens_apply:alias_tokens_apply E G)
  (ndef:S[m][E[a]] <> C_Unalloc) mv `(mv = S[m][E[a] <- C (E[x])]):
  exists mt, stateConsistent (ET, LT, ST[m <- None][m' <- Some mt], F) s (E, L, S[m' <- mv], U, b) rt.
Proof.
invc types. invc stmt_types. exists mt. apply SOT; try assumption.
- assert (domain (S[m][E[a] <- C (E[x])]) ⊆ domain (S[m])).
  { apply domain_update_incl. invc mems_type. rewrite mems_default. assumption. }
  eapply memMoreTokensTypes, memRenameTypes; crush.
- eapply oracleUpdateMem with (mv1 := merge_all_mems ST S); try eassumption.
  + symmetry. eapply merge_all_mems_rename_update; crush.
  + reflexivity.
  + eapply env_types. assumption.
Qed.

Lemma freePreservation ET LT ST F s E L S U G m' m a b rt
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtFree G m' m a b) rt)
  (tokens_apply:alias_tokens_apply E G)
  (ndef:S[m][E[a]] <> C_Unalloc) mv `(mv = S[m][E[a] <- C_Unalloc]):
  exists mt, stateConsistent (ET, LT, ST[m <- None][m' <- Some mt], F) s (E, L, S[m' <- mv], U, b) rt.
Proof.
invc types. invc stmt_types. exists mt. apply SOT; try assumption.
- assert (domain (S[m][E[a] <- C_Unalloc]) ⊆ domain (S[m])).
  { apply domain_update_incl. invc mems_type. rewrite mems_default. assumption. }
  eapply memMoreTokensTypes, memRenameTypes; crush.
- eapply oracleFreeInMem with (mv1 := merge_all_mems ST S); try eassumption.
  + symmetry. eapply merge_all_mems_rename_update; crush.
Qed.

Lemma splitPreservation ET LT ST F s E L S U G m' m'' m A b rt
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtSplit G m' m'' m A b) rt)
  (tokens_apply:alias_tokens_apply E G)
  mv' `(mv' = restrict_mem (S[m]) (minus (domain (S[m])) (lookup_set E A)))
  mv'' `(mv'' = restrict_mem (S[m]) (lookup_set E A)):
  exists mt F', stateConsistent (ET, LT, ST[m <- None][m' <- Some mt][m'' <- Some (MemFocus A)], F ∪ F') s (E, L, S[m' <- mv'][m'' <- mv''], U, b) rt.
Proof.
invc types. invc stmt_types.
- exists (MemFocus (mA\A )), BSet.empty. rewrite union_comm, empty_neutral_union. split; try assumption.
  + eapply memSplitFocusTypes; crush.
  + eapply oracleSameMem; [|eassumption]. eapply merge_all_mems_split; crush.
- exists MemPanorama, A. split; try assumption.
  + eapply memSplitPanoramaTypes; crush.
  + eapply oracleSameMem; [|eassumption]. eapply merge_all_mems_split; crush.
Qed.

Lemma letPreservation ET LT ST F s E L S U G Z m (t b:stmt) (rt:ty)
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtLet G Z m t b) rt):
  exists B, stateConsistent (ET, B::LT, ST, F) s (E, (blockI E Z m t)::L, S, U, b) rt.
Proof.
invc types. invc stmt_types. exists (TB ET' ZT rt0). split; try assumption.
- eapply labenvOfType_upd.
  + eapply envOfType_weakening; crush.
  + assumption.
  + assumption.
- eapply memMoreTokensTypes; crush.
- eapply oracleAddLab; crush.
Qed.

Lemma gotoPreservation ET LT ST F s E L S U G l Y m rt
  (types:stateConsistent (ET, LT, ST, F) s (E, L, S, U, stmtGoto G l Y m) rt)
  (tokens_apply:alias_tokens_apply E G)
  blk (Ldef:get L (counted l) blk)
  E' (updOk:updE (block_E blk) E (block_Z blk) Y  = E')
  S' `(empty[(block_m blk) <- S[m]] = S'):
  exists ET', stateConsistent (ET', drop (counted l) LT, ompty[(block_m blk) <- Some MemPanorama], BSet.empty) s (E', drop (counted l) L, S', U, block_s blk) rt.
Proof.
invc types. invc stmt_types. destruct blk0 as [ET' ZT rt]. destruct blk as [E' Z m' b]. simpl in *.
exists (updET ET' ZT Z).
destruct (labenv_types _ _ _ _ get0) as [blk [get1 [stmt_types env'_types]]]. assert (blkEQ := get_functional Ldef get1). subst blk. apply SOT; simpl in *.
- assumption.
- eapply envOfType_listUpdate; crush.
- apply labenvOfType_drop; crush.
- invc mems_type. eapply memPanoramaOnlyTypes; crush.
- eapply oracleDropLab, oracleChangeEnv, oracleSameMem with (mv1 := merge_all_mems ST S);
  eauto using merge_all_mems_one_mem, focusEmpty_memsEmpty.
Qed.

(* Theorem 1: Type consistency is preserved by doing a step *)
Theorem preservation (typeEnv:type_env) s (st:state) ty
  (tokens_apply:alias_tokens_apply (state_E st) (state_G st))
  (types:stateConsistent typeEnv s st ty) (st':state) (step: fstep st st'):
  exists typeEnv' s', stateConsistent typeEnv' s' st' ty.
Proof.
destruct typeEnv as [[[ET LT] ST] F]. destruct st as [[[[E L] S] U] b].
invc step.
(* Assignment *)
- edestruct (expPreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* If true *)
- edestruct (ifPreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* If false *)
- edestruct (ifPreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* goto *)
- edestruct (gotoPreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* let *)
- edestruct (letPreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* alloc *)
- edestruct (allocPreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* free *)
- edestruct (freePreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* store *)
- edestruct (storePreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* load *)
- edestruct (loadPreservation types); try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* merge *)
- edestruct (mergePreservation types) as [mt [A H]]; try (reflexivity || eassumption).
  eexists. eexists. eassumption.
(* split *)
- edestruct (splitPreservation types) as [mt [F' H]]; try (reflexivity || eassumption).
  eexists. eexists. eassumption.
Qed.

(* Type consistency is preserved by performing an arbitrary amount of steps *)
Lemma preservationStar (typeEnv:type_env) s (st:state) rt
  (tokens_correct: alias_tokens_correct typeEnv (state_b st) rt)
  (types:stateConsistent typeEnv s st rt) (st':state) (steps: fstepStar st st'):
  exists typeState' s', stateConsistent typeState' s' st' rt.
Proof.
(* We need to generalize the induction statement, to ensure the parameters of tokens_correct do not change:
   Given an alias-correct program, the statements which can be reached are not necessarily alias-correct
   as there could be executions which cannot be reached from the initial program (think, for example,
   of the branches of an "if"). *)
remember st as stIN. remember typeEnv as typeEnvIN. remember s as sIN.
assert (typesIN := types). rewrite HeqstIN in types, steps.
rewrite HeqtypeEnvIN in types. rewrite HeqsIN in types.
assert (stepsIN: fstepStar stIN st). { rewrite HeqstIN. apply fstepStarRefl. }
destruct stIN as [[[[E L] S] U] b]. specialize (tokens_correct _ _ _ _ _ typesIN). simpl in *.
clear- stepsIN steps types tokens_correct. revert typeEnv s types.
(* now the induction statement is general enough *)
induction steps; intros.
- exists typeEnv. exists s. assumption.
- destruct st1 as [[[[E1 L1] S1] U1] b1]. simpl in *.
  edestruct preservation with (st:=(E1, L1, S1, U1, b1)) (st':=st2) as [typeEnv2 [s2 P]]; try eassumption.
  + eapply tokens_correct; crush.
  + eapply IHsteps; try eassumption. eapply fstepStarTrans.
    * eassumption.
    * apply fstep_fstepStar. assumption.
Qed.

(* Well-typed programs can only terminate with variables of the proper type *)
Lemma evalsTo_type (typeEnv:type_env) s (st:state) rt
  (tokens_correct: alias_tokens_correct typeEnv (state_b st) rt)
  (types: stateConsistent typeEnv s st rt) v (evals:fevalsTo st v):
  valOfType v rt.
Proof.
apply fevalsTo_star in evals. destruct evals as [E [L [S [U [G [x [steps EQ]]]]]]].
edestruct preservationStar with (st:=st) (st':=(E, L, S, U, stmtReturn G x)) as [typeEnv2 [s2 P]]; try eassumption.
- invc P. eapply env_types. invc stmt_types. assumption.
Qed.


(*Print Assumptions evalsTo_type.*)


(* 
*** Local Variables: ***
*** coq-load-path: ("infra" "constr" "il" "isa" ".") ***
*** End: ***
*)
