Require Import Le Plus List Setoid Coq.Arith.Compare Coq.Program.Tactics.
Require Import EnvTy ILM.

Open Scope map_scope.

Set Implicit Arguments.

(* This file contains the formalisation of the type system, and the notion of type-consistent states *)
(* It roughly corresponds to Sections 4.2 and 5.1, except for the proof of type preservation which can be found in ILM_Preservation.v *)

(** Some basics and tactics  *)

(* Attempt to fix annoying <> overwrite, and use * more easily to construct tuple types *)
Notation "a <> b" := (a <> b)%type.
Notation "a * b" := (a * b)%type.

(* From the Coq library, but fixed to work better *)
Ltac autoinjection tac :=
  match goal with
    | [ H : ?f ?a = ?f ?a' |- _ ] => tac H
  end.
Ltac inject H := progress (inversion H ; subst*; clear_dups) ; clear H.
Ltac autoinjections := repeat (clear_dups ; subst*; autoinjection ltac:inject).

(* simple automatism and other useful tactics *)
Ltac discriminate_eq := solve [ match goal with
    | [ H: ?x = _, I: ?x = _ |- _ ] => rewrite H in I; discriminate
    | [ H: ?x = _, I: _ = ?x |- _ ] => rewrite H in I; discriminate end ].
Ltac destruct_lookup :=
  match goal with
    | [ x1: _, x2: _ |- context[@lookup _ _ _ _ (@update _ _ _ _ ?m ?x1 ?val) ?x2] ]
      => destruct_prop(x1=x2); simplify lookup'
    | [x1: _, x2: _, H : context[@lookup _ _ _ _ (@update _ _ _ _ ?m ?x1 ?val) ?x2] |- _ ]
      => destruct_prop(x1=x2); simplify lookup'
  end.
Ltac crush := now repeat (
  discriminate || discriminate_eq || (* simple "finish" tatics *)
  intros || destruct_conjs || autoinjections || simplify lookup' || (* simple improvement tactics *)
  (progress eauto) || (* the core *)
  split || destruct_lookup || (* simple case distinctions *)
  congruence || tauto || (* complex finish tactics *)
  cbv in * || autorewrite with core in * ). (* complex rewriting *)


Ltac destruct_lookup_a := destruct_lookup; [crush|].
Ltac destruct_lookup_b := destruct_lookup; [|crush].
Tactic Notation "dupl_apply" ident(HimplY) "in" ident(H):=
let new := fresh H in
  assert(new:=H);
  apply HimplY in new.



(* Useful fatcs about domains of maps *)
Lemma domain_empty {W} {X} `{EqDec W eq} (def:X): domain (@BMap.empty W X def) = BSet.empty.
Proof.
cset_tac.
- eapply domain_spec in H0. now eauto. simplify lookup. rewrite default_spec; now eauto.
- isabsurd.
Qed.

Lemma domain_spec2: forall (W : Type) `{EqDec W eq} X (m : map W X) (x : W),
      x ∈ (domain m) <-> lookup m x <> default m.
Proof.
intros. split; intros.
- rewrite <- domain_spec. now eauto.
- rewrite <- negb_involutive. apply negb_prop_intro. intros A. apply negb_prop_elim in A. apply H0.
  apply domain_spec. exact A.
Qed.

Lemma onv_domain X (m: onv X) `(default m = None):
  forall x, x ∈ (domain m) -> exists v, m[x] = Some v.
Proof.
intros. case_eq (m[x]); intros.
- exists x0. reflexivity.
- exfalso. eapply (@domain_spec2 var). now eexact H0. rewrite H. exact H1.
Qed.


Lemma domain_update {W} {X} `{EqDec W eq} (m: map W X) (k: W) (x: X) `(x <> default m):
domain (m[k <- x]) = domain m ∪ {{k}}.
Proof.
eapply set_extensionality. intros. eapply bool_extensionality; intros. 
  cset_tac. eapply domain_spec2 in H1. destruct_lookup. right. left. now eauto.
    left. eapply domain_spec2. rewrite default_spec2 in H1. exact H1.
  cset_tac. destruct H1 as [A|[A|A]].
    eapply domain_spec2. destruct_lookup. rewrite default_spec2. exact H0.
      eapply domain_spec2 in A. rewrite default_spec2. exact A.
    eapply domain_spec2. simplify lookup. rewrite default_spec2. exact H0.
    contradiction A.
Qed.

Lemma domain_update_incl {W} {X} `{EqDec W eq} (m: map W X) (k: W) (x: X) `(m[k] <> default m):
domain (m[k <- x]) ⊆ domain m.
Proof.
intros y IN. rewrite domain_spec2. rewrite domain_spec2 in IN. destruct_lookup.
+ assumption.
+ rewrite default_spec2 in *. assumption.
Qed.

Lemma lookup_set_update {W} {X} `{EqDec W eq} (m:map W X) (k:W) (x:X) (A:cset W) `(k ∉ A):
lookup_set (m[k <- x]) A = lookup_set m A.
Proof.
cset_tac; rewrite lookup_set_spec in *.
- destruct H1 as [y [IN EQ]]. subst x0. exists y. split.
  + assumption.
  + destruct_lookup_a. reflexivity.
- destruct H1 as [y [IN EQ]]. subst x0. exists y. split; crush.
Qed.

(* Some set facts *)
Lemma union_empty_l X (A: cset X): BSet.empty ∪ A = A.
Proof.
cset_tac. destruct H; now eauto. right. exact H.
Qed.
Lemma union_empty_r X (A: cset X): A ∪ BSet.empty = A.
Proof.
cset_tac. destruct H; now eauto. left. exact H.
Qed.
Lemma element_single X (m:X) (x:X): x ∈ {{m}} <-> x=m.
Proof.
split.
intros. cset_tac. destruct H; crush.
intros. subst. cset_tac. left. reflexivity.
Qed.
Lemma element_fromList X (x:X) (s:cset X): x ∈ fromList (elements s) = x ∈ s.
Proof.
apply bool_extensionality; rewrite in_fromList, elements_spec; tauto.
Qed.
Lemma incl_union2 X (A:cset X) B: A ⊆ A ∪ B.
Proof.
rewrite union_comm. apply incl_union.
Qed.
Lemma incl_union_l X (A: cset X) B C: A ⊆ B -> A ⊆ B ∪ C.
Proof.
intros. intros x IN. cset_tac. left. apply H. assumption.
Qed.
Lemma incl_union_r X (A B C: cset X): A ⊆ C -> A ⊆ B ∪ C.
Proof.
rewrite union_comm. now apply incl_union_l.
Qed.
Lemma union_incl_split X (A B C:cset X) : A ⊆ C -> B ⊆ C -> A ∪ B ⊆ C.
Proof.
intros IN1 IN2 x. cset_tac. destruct H; now auto.
Qed.
Lemma incl_minus_swap X (A B C:cset X): (A ∪ B) \ C ⊆ (A \ C) ∪ B.
intros x IN. cset_tac. tauto.
Qed.

Hint Resolve union_comm incl_union incl_union2 incl_union_r incl_union_l union_incl_split minus_incl.
Hint Rewrite minus_empty BSet.empty_spec default_spec default_spec2.

(* Some list facts *)
Lemma length_drop (X: Type) (L1 L2: list X) n: length L1 = length L2 -> length (drop n L1) = length (drop n L2).
Proof.
revert L1 L2. induction n; simpl; intros.
- assumption.
- apply IHn. destruct L1, L2; simpl in *; crush.
Qed.

Lemma drop_nil X n: drop n (@nil X) = nil.
Proof.
induction n; simpl; eauto.
Qed.

Lemma drop_app_snd X (l1 l1' l2: list X) n: length l1 = length l1' ->
  length l1 <= n -> drop n (l1++l2) = drop n (l1'++l2).
Proof.
revert l1 l1' l2. induction n; intros ? ? ? EQ LE.
- inversion LE. rewrite H0. destruct l1, l1'; try discriminate. simpl. reflexivity.
- destruct l1, l1'; simpl in *; try reflexivity; try discriminate. eapply IHn; omega.
Qed.

Lemma drop_app_fst X (l1 l2: list X) n:
  length l1 > n -> drop n (l1++l2) = (drop n l1)++l2.
Proof.
revert l1 l2. induction n; intros ? ? GT.
- simpl. reflexivity.
- destruct l1; simpl in *.
  + inversion GT.
  + eapply IHn. omega.
Qed.

Lemma get_app_snd X (l1 l1' l2: list X) n x: length l1 = length l1' -> length l1 <= n ->
  get (l1++l2) n x -> get (l1'++l2) n x.
Proof.
revert l1 l1' l2 x. induction n; intros ? ? ? ? EQ LE.
- inversion LE. clear LE. destruct l1, l1'; simpl in *; try discriminate. tauto.
- destruct l1, l1'; simpl in *; [tauto|discriminate|discriminate|]. intros GET. inv GET. apply getLS.
  apply IHn with (l1:=l1); try assumption; omega.
Qed.

(* Alternative characterisation of disjointness of memories *)
Definition disjoint X (A: cset X) (B: cset X):Prop := forall x:X, ~(x ∈ A /\ x ∈ B).
Hint Unfold disjoint.

Lemma disjoint_in_one X A B (x:X): disjoint A B -> x ∈ A -> x ∉ B.
Proof.
intros. intros H1. unfold disjoint in H. eapply H. split; eassumption.
Qed.

Lemma disjoint_empty X (A:cset X): disjoint A (BSet.empty).
Proof.
intros x [IN1 IN2]. rewrite BSet.empty_spec in IN2. now isabsurd. Qed.

Lemma disjoint_sym X (A:cset X) B: disjoint A B -> disjoint B A.
Proof.
unfold disjoint. intros. intros [H1' H2']. eapply H. split; eassumption.
Qed.
Hint Resolve disjoint_sym.

Lemma disjoint_minus X (A:cset X) B: disjoint A (B\A).
Proof.
intros x [IN1 IN2]. apply minus_in_in in IN2. tauto. Qed.

Lemma disjoint_equiv_meet_empty X (A: cset X) B: disjoint A B <-> A ∩ B = BSet.empty.
Proof.
unfold disjoint. split; intros.
  cset_tac. eapply H. split; eassumption. isabsurd. isabsurd.
  intros [H' H'']. assert (x ∈ (A ∩ B)) by (cset_tac; crush). rewrite H in H0. crush.
Qed.

Lemma disjoint_split X (A:cset X) B C: disjoint A B -> disjoint A C -> disjoint A (B ∪ C).
Proof.
intros D1 D2 x [IN1 IN2].
cset_tac. destruct IN2.
- eapply D1; crush.
- eapply D2; crush.
Qed.

Lemma disjoint_incl_lr X (C D A B:cset X): A ⊆ C -> B ⊆ D -> disjoint C D -> disjoint A B.
Proof.
intros INC1 INC2 DISJ x [IN1 IN2]. eapply DISJ; crush.
Qed.
Lemma disjoint_incl_r X (A:cset X) B C: B ⊆ C -> disjoint A C -> disjoint A B.
Proof.
eapply disjoint_incl_lr; eauto using incl_refl.
Qed.

(* Functions dealing with pointer types *)
Definition isPtr (t:ty) := match isPtr_sigma t with inleft _ => true | inright _ => false end.
Lemma ptrIsPtr: forall t:ty, isPtr (ptrTo t).
Proof.
intros. unfold isPtr. destruct (isPtr_sigma (ptrTo t)).
- constructor.
- apply (n t). reflexivity.
Qed.


(** Alias tokens *)
Inductive alias_token: Type :=
| AliasEq: forall a1 a2:var, alias_token
| AliasNeq: forall a1 a2:var, alias_token.

Definition alias_tokens := cset alias_token.

Inductive alias (G: alias_tokens): alias_token -> Prop := (* Figure 4.1 *)
| AliasEqRefl: forall a, alias G (AliasEq a a)
| AliasEqSym: forall a1 a2, alias G (AliasEq a2 a1) -> alias G (AliasEq a1 a2)
| AliasEqTrans: forall a1 a2 a3, alias G (AliasEq a1 a3) -> alias G (AliasEq a3 a2) -> alias G (AliasEq a1 a2)
| AliasNeqSym: forall a1 a2, alias G (AliasNeq a2 a1) -> alias G (AliasNeq a1 a2)
| AliasEqNeq: forall a1 a2 a3, alias G (AliasEq a1 a3) -> alias G (AliasNeq a3 a2) -> alias G (AliasNeq a1 a2)
| AliasAss: forall t, t ∈ G -> alias G t.

Definition alias_token_applies (E: env val) (t: alias_token): Prop := (* Definition 15 (E satisfies the token <=> the token applies to E) *)
match t with
| AliasEq a1 a2 => E[a1] = E[a2]
| AliasNeq a1 a2 => E[a1] <> E[a2] end.


Definition alias_tokens_apply (E: env val) (G: alias_tokens): Prop := (* Definition 15 *)
  forall t, t ∈ G -> alias_token_applies E t.

(* Soundness of alias token derivation - Lemma 1 *)
Lemma alias_sound G E (apply: alias_tokens_apply E G) t (derive: alias G t): alias_token_applies E t.
Proof.
induction derive; simpl in *|-*.
- reflexivity.
- symmetry. exact IHderive.
- rewrite IHderive1. exact IHderive2.
- intros A. apply IHderive. symmetry. exact A.
- rewrite IHderive1. exact IHderive2.
- apply apply. assumption.
Qed.

(* Some mroe useful facts *)
Lemma alias_incl G1 G2: G1 ⊆ G2 -> forall t, alias G1 t -> alias G2 t.
Proof.
intros INCL t AL. induction AL.
- eapply AliasEqRefl; crush.
- eapply AliasEqSym; crush.
- eapply AliasEqTrans; crush.
- eapply AliasNeqSym; crush.
- eapply AliasEqNeq; crush.
- eapply AliasAss; crush.
Qed.

Lemma alias_tokens_apply_incl E G1 G2: G2 ⊆ G1 -> alias_tokens_apply E G1 -> alias_tokens_apply E G2.
Proof.
intros. intros t IN. crush.
Qed.


Definition alias_disjoint G A1 A2:Prop := forall a1 a2, a1 ∈ A1 -> a2 ∈ A2 -> alias G (AliasNeq a1 a2). (* Definition 9 *)
Hint Unfold alias_disjoint.

Lemma alias_disjoint_sym G A1 A2: alias_disjoint G A1 A2 <-> alias_disjoint G A2 A1.
Proof.
unfold alias_disjoint. split; intros; eapply AliasNeqSym; crush.
Qed.
Hint Resolve alias_disjoint_sym.

Lemma alias_disjoint_incl_r G A B1 B2: B1 ⊆ B2 -> alias_disjoint G A B2 -> alias_disjoint G A B1.
Proof.
unfold alias_disjoint. crush.
Qed.

Lemma alias_disjoint_split G A B1 B2:
  alias_disjoint G A B1 -> alias_disjoint G A B2 -> alias_disjoint G A (B1 ∪ B2).
Proof.
unfold alias_disjoint. intros. cset_tac. destruct H2; crush.
Qed.

Lemma alias_disjoint_lookup_disjoint A1 A2 G E: (* Lemma 1 *)
  alias_disjoint G A1 A2 -> alias_tokens_apply E G -> disjoint (lookup_set E A1) (lookup_set E A2).
Proof.
intros disj apply x [IN1 IN2]. rewrite lookup_set_spec in *. destruct_conjs.
specialize (disj _ _ H1 H). specialize (alias_sound apply disj). simpl. subst x. symmetry in H0.
tauto.
Qed.





(** In the following, we use alias-annoated programs *)
Definition stmt := @ILM.stmt alias_tokens.
Definition state := @ILM.state alias_tokens.
Definition labenv := @ILM.labenv alias_tokens.
Definition empty_state := @ILM.empty_state alias_tokens.
Definition state_G (st:state) := stmt_get_ann (state_b st).
Hint Unfold state_G empty_state labenv state stmt.


(** Memory types *)
Inductive memory_type: Type :=
| MemFocus: cset var -> memory_type
| MemPanorama.

(* This is accessibility - Definition 8 *)
Definition in_mem (G: alias_tokens) (F: cset var) (t: memory_type) (a: var): Prop := match t with
| MemFocus A => exists a':var, a' ∈ A /\ alias G (AliasEq a a')
| MemPanorama => forall a', a' ∈ F -> alias G (AliasNeq a a') end.

(* Some useful facts about accessibility and memories in general *)
Lemma in_mem_incl_G G1 G2 F t a: G1 ⊆ G2 -> in_mem G1 F t a -> in_mem G2 F t a.
Proof.
intros INCL inmem. destruct t; simpl in *.
- destruct inmem as [a' [IN AEQ]]. exists a'. split; [assumption|].
  eapply alias_incl; eassumption.
- intros a' IN. eapply alias_incl; eauto.
Qed.
Hint Resolve in_mem_incl_G.

Lemma merge_mem_domain (m1 m2: mem) (def1: default m1 = C_Unalloc) (def2: default m2 = C_Unalloc):
domain (merge_mem m1 m2) = domain m1 ∪ domain m2.
Proof.
unfold merge_mem. cset_tac; specialize (mapjoin_spec m1 m2 (domain m1) x); intros [INM1 INM2].
- rewrite domain_spec2, mapjoin_default in H. destruct_prop(x ∈ domain m1).
  + left. assumption.
  + right.  rewrite domain_spec2. apply INM2 in n. rewrite n in *. assumption.
- rewrite domain_spec2, mapjoin_default. destruct_prop(x ∈ domain m1).
  + dupl_apply INM1 in i. rewrite i0. rewrite domain_spec2 in i. rewrite def2, <- def1. assumption.
  + dupl_apply INM2 in n. rewrite n0. rewrite 2!domain_spec2 in *. tauto.
Qed.

Lemma merge_mem_default (m1 m2: mem): default (merge_mem m1 m2) = default m2.
Proof.
unfold merge_mem. rewrite mapjoin_default. reflexivity.
Qed.

Lemma merge_mem_equiv (m1 m2 m1' m2': mem) (samedef: default m1 = default m1'):
  m1 === m1' -> m2 === m2' -> merge_mem m1 m2 === merge_mem m1' m2'.
Proof.
intros EQM1 EQM2.
assert (samedom:domain m1 = domain m1'). { cset_tac; rewrite domain_spec2 in *; congruence. }
intros x. unfold merge_mem. rewrite <-samedom. destruct_prop (x ∈ (domain m1)).
- transitivity (m1[x]); eapply mapjoin_spec in i; rewrite i; crush.
- transitivity (m2[x]); eapply mapjoin_spec in n; rewrite n; crush.
Qed.


Lemma restrict_mem_default (m:mem) c:
default (restrict_mem m c) = C_Unalloc.
Proof.
unfold restrict_mem. induction (elements c); simpl.
- unfold mempty. rewrite default_spec. reflexivity.
- rewrite default_spec2. assumption.
Qed.

Lemma restrict_mem_domain (m:mem) c (def:default m = C_Unalloc):
domain (restrict_mem m c) = domain m ∩ c.
Proof.
assert (def_restr := restrict_mem_default m c).
unfold restrict_mem, mem in *. apply set_extensionality. intros. rewrite intersection_spec.
apply bool_extensionality; intros H; [bool_to_prop|bool_to_prop in H]; rewrite !domain_spec2, !def_restr in *.
- destruct_prop(x ∈ fromList (elements c)).
  + split. 
    * eapply update_list_upd with (f:=lookup m) in i. erewrite <-i, def. eassumption.
    * rewrite <- element_fromList. assumption.
  + exfalso. eapply update_list_no_upd with (f:=lookup m) in n. erewrite n in H. apply H. unfold mempty. crush.
- destruct H as [NDEF INC]. rewrite <-element_fromList in INC. eapply update_list_upd with (f:=lookup m) in INC.
  erewrite INC, <-def. assumption. 
Qed.

Lemma restrict_mem_equiv (m1 m2: mem) A:
  m1 === m2 -> restrict_mem m1 A === restrict_mem m2 A.
Proof.
intros EQM x. unfold restrict_mem. induction (elements A); simpl.
- reflexivity.
- destruct_lookup.
  + rewrite EQM. reflexivity.
  + apply IHl.
Qed.

Lemma disjoint_restrict (mv1 mv2: mem) A (def: default mv2 = C_Unalloc):
  disjoint (domain mv1) (domain mv2) -> disjoint (domain mv1) (domain (restrict_mem mv2 A)).
Proof.
intros disj. rewrite restrict_mem_domain by assumption. eapply disjoint_incl_r with (C:= domain mv2).
- eapply meet_incl, incl_refl.
- assumption.
Qed.

(* Validity of memory types - this is condition 1 of type-consistency (Definition 14) *)
(* We also need the "ST_default" condition for technical reasons. maps do not have a fixed "default"
   element like the environments in the thesis do, so we have to remember which element was the
   default element *)
Inductive memTypesValid (ET:onv ty) (ST:onv memory_type) (G:alias_tokens) (F:cset var): Prop :=
  MTV
  (ST_default: default ST = None)
  (focus_sane: forall a, a ∈ F -> exists t, ET[a] = Some (ptrTo t)) (* necessary to justify that setting a previously unset variable does not change the domain specifications *)
  (unique_mtop: exists! m, ST[m] = Some MemPanorama)
  (focus_disjoint:forall m1 A1 m2 A2, ST[m1] = Some (MemFocus A1) -> ST[m2] = Some (MemFocus A2)
            -> m1 <> m2 -> alias_disjoint G A1 A2)
  (mems_in_focus:forall m A, ST[m] = Some (MemFocus A) ->  A ⊆ F)
  (focus_in_mem:forall a, a ∈ F -> exists m A, ST[m] = Some (MemFocus A) /\ a ∈ A)
  :memTypesValid ET ST G F.

(* Some useful facts about valid memory types *)
Lemma mtop_unique (ET:onv ty) (ST:onv memory_type) G F: memTypesValid ET ST G F ->
  forall m1 m2, ST[m1] = Some MemPanorama -> ST[m2] = Some MemPanorama -> m1 = m2.
Proof.
intros types_valid. invc types_valid. intros. destruct unique_mtop as [mtop [EX UN]].
transitivity mtop.
- symmetry. crush.
- crush.
Qed.

Lemma disjointF_disjointAll ET ST G F A (valid:memTypesValid ET ST G F) (disj:alias_disjoint G A F):
forall m A', ST[m] = Some (MemFocus A') -> alias_disjoint G A' A.
Proof.
intros m A' EQ. inv valid.
assert (A' ⊆ F) by eauto using mems_in_focus.
apply alias_disjoint_sym.
crush.
Qed.
Lemma focus_disjoint_rest ET ST G F m A (valid:memTypesValid ET ST G F):
  ST[m] = Some (MemFocus A) -> alias_disjoint G A (F\A).
Proof.
intros EQ. invc valid. intros ? ? IN1 IN2. cset_tac.
destruct (focus_in_mem _ H) as [m' [A' [EQ' IN']]]. destruct_prop (m'=m).
- exfalso. apply H0. subst m'. rewrite EQ in EQ'. autoinjections. assumption.
- eapply focus_disjoint with (m1 := m) (m2 := m'); crush.
Qed.


(* Having just a panorama memory is valid *)
Lemma memPanoramaOnlyValid ET G m: memTypesValid ET (ompty[m <- Some MemPanorama]) G BSet.empty.
Proof.
apply MTV; unfold ompty in *; unfold Env.empty in *.
- crush.
- intros a X. crush.
- exists m. split; crush.
- intros. destruct_lookup; discriminate.
- intros. destruct_lookup; discriminate.
- intros. crush.
Qed.

(* the type is a statement about the domain of the memory, and it makes some proofs easier to actually
   define it that way *)
Definition memoryOfType mt F E (d:cset val):Prop := match mt with
| MemFocus A => d ⊆ lookup_set E A
| MemPanorama => disjoint (d) (lookup_set E F) end.

(* Well-typedness of memories: Condition 5 of type-consistency (Definition 14) *)
(* It also includes memory type validity since we often need them together *)
Inductive memoriesOfType (ET:onv ty) (ST:onv memory_type) (G:alias_tokens) (F:cset var) (E:env val) (S:env mem) (U:cset val):Prop :=
  MOT
  (types_valid:memTypesValid ET ST G F)
  (mems_default: forall m, default (S[m]) = C_Unalloc)
  (mems_in_U:forall m mt, ST[m] = Some mt -> domain (S[m]) ⊆ U) (* used by alloc to justify updating the oracle *)
  (mems_type:forall m mt, ST[m] = Some mt -> memoryOfType mt F E (domain (S[m])))
  :memoriesOfType ET ST G F E S U.

(* Some useful facts about well-typed memories *)

(* Lemma 2: Two different well-typed memories are disjoint *)
Lemma welltypedMemoriesAreDisjoint ET (ST:onv memory_type) G F E S U: memoriesOfType ET ST G F E S U ->
  alias_tokens_apply E G -> forall m1 t1 m2 t2, ST[m1] = Some t1 -> ST[m2] = Some t2 ->
  m1 <> m2 -> disjoint (domain (S[m1])) (domain (S[m2])).
Proof.
intros types tokens_apply m1 t1 m2 t2 m1ty m2ty ineq.
inv types. inv types_valid.
destruct t1; destruct t2; dupl_apply mems_type in m1ty; dupl_apply mems_type in m2ty; simpl in *|-*.
- assert(alias_disjoint G c c0) by eauto. apply (disjoint_incl_lr m1ty0 m2ty0).
  eapply alias_disjoint_lookup_disjoint; crush.
- assert (domain (S[m1]) ⊆ lookup_set E F) by eauto using incl_trans, lookup_set_incl, mems_in_focus.
  eapply disjoint_sym, disjoint_incl_r; crush.
- assert (domain (S[m2]) ⊆ lookup_set E F) by eauto using incl_trans, lookup_set_incl, mems_in_focus.
  eauto using disjoint_incl_lr, incl_refl.
- exfalso.  apply ineq. eapply mtop_unique; crush.
Qed.

(* Lemma 3: If a variable is accessible in a memory, then its value is never contained
   in another memory *)
Lemma in_mem_not_in_other (ET:onv ty) (ST:onv memory_type) G F E S U m1 t1 a
  (types:memoriesOfType ET ST G F E S U) (tokens_apply: alias_tokens_apply E G)
  (ty1:ST[m1] = Some t1) (inmem: in_mem G F t1 a):
  forall v m2 t2 (ineq: m1 <> m2) (ty2:ST[m2] = Some t2) (val:E[a] = v), v ∉ domain (S[m2]).
Proof.
intros. inv types. inv types_valid. intros IN. dupl_apply mems_type in ty2.
destruct t1, t2; simpl in *.
- apply ty0 in IN. destruct inmem as [a' [IN' AL']]. eapply (alias_sound tokens_apply) in AL'. simpl in *.
  assert (E[a'] ∈ lookup_set E c). { rewrite lookup_set_spec. exists a'. crush. }
  eapply (@alias_disjoint_lookup_disjoint c0 c); crush.
- destruct inmem as [a' [IN' AL']]. apply (ty0 (E[a])). split.
  + crush.
  + rewrite lookup_set_spec. exists a'. split.
    * eapply mems_in_focus; crush.
    * symmetry. eapply (alias_sound tokens_apply AL').
- dupl_apply ty0 in IN. rewrite lookup_set_spec in *. destruct IN0 as [x [INx EQx]].
  eapply (mems_in_focus _ _ ty2) in INx. eapply (alias_sound tokens_apply (inmem _ INx)). crush.
- apply ineq. eapply mtop_unique; crush.
Qed.

Lemma allocInNoMem ET ST G F E S U m mt: memoriesOfType ET ST G F E S U ->
  ST[m] = Some mt -> allocator U ∉ domain (S[m]).
Proof.
intros types hasty IN. invc types. specialize (mems_in_U _ _ hasty). eapply mems_in_U in IN.
eapply allocatorGivesNew. eassumption.
Qed.

Lemma focusEmpty_memsEmpty ET (ST: onv memory_type) G F E S U m
  (fempty: F = BSet.empty) (ispan: ST[m] = Some MemPanorama):
  memoriesOfType ET ST G F E S U -> forall m' mt' v, m' <> m -> ST[m'] = Some mt' -> S[m'][v] = C_Unalloc.
Proof.
intros mems_type. intros. invc mems_type. invc types_valid. destruct mt'.
+ specialize (mems_type0 _ _ H0). simpl in *. specialize (mems_in_focus _ _ H0). rewrite <- (mems_default m').
  rewrite <- domain_spec. intros IN. apply mems_type0 in IN. rewrite lookup_set_spec in IN.
  destruct IN as [x [IN _]]. apply mems_in_focus in IN. crush.
+ exfalso. destruct unique_mtop as [mtop [ISM UNI]]. apply H. transitivity mtop; [symmetry|]; now apply UNI.
Qed.


(* a memory merged from two focused ones is well-typed *)
Lemma mergedFocusedMemTypes mv1 A1 mv2 A2 F E:
   default mv1 = C_Unalloc -> default mv2 = C_Unalloc ->
   memoryOfType (MemFocus A1) F E (domain mv1) -> memoryOfType (MemFocus A2) F E (domain mv2) ->
   memoryOfType (MemFocus (A1 ∪ A2)) F E (domain (merge_mem mv1 mv2)).
Proof.
intros def1 def2 types1 types2. rewrite merge_mem_domain by crush. simpl in *. rewrite lookup_set_union.
apply incl_union_lr; crush.
Qed.

(* a focused memory merged into the all-memory is well-typed *)
Lemma mergedPanoramaTypes mv1 A1 mv2 F E G
  (apply:alias_tokens_apply E G) (disj: alias_disjoint G A1 (F\A1)):
  default mv1 = C_Unalloc -> default mv2 = C_Unalloc ->
  memoryOfType (MemFocus A1) F E (domain mv1) -> memoryOfType MemPanorama F E (domain mv2) ->
  memoryOfType MemPanorama (F\A1) E (domain (merge_mem mv1 mv2)).
Proof.
intros def1 def2 types1 types2. rewrite merge_mem_domain by crush. simpl in *.
intros x [IN1 IN2]. cset_tac. destruct IN1 as [IN1|IN1].
- eapply (@alias_disjoint_lookup_disjoint A1 (F\A1) G); eauto.
- apply (types2 x). split.
  + assumption.
  + now eapply lookup_set_minus with (t:=A1).
Qed.


(* a memory split-off from another one is well-typed *)
Lemma splitMemoryTypes (mv:mem) F E A:
  default mv = C_Unalloc -> 
  memoryOfType (MemFocus A) F E (domain (restrict_mem mv (lookup_set E A))).
Proof.
intros def. rewrite restrict_mem_domain by crush. simpl.
rewrite meet_comm. apply meet_incl, incl_refl.
Qed.

(* the leftover after a split from a focus memory is well-typed *)
Lemma splitFocusRemainderTypes (mv:mem) A1 F E A:
  default mv = C_Unalloc -> 
  memoryOfType (MemFocus A1) F E (domain mv) ->
  memoryOfType (MemFocus (A1\A)) F E (domain (restrict_mem mv (domain mv \ lookup_set E A))).
Proof.
intros def types. rewrite restrict_mem_domain by crush. simpl in *.
rewrite meet_comm. apply meet_incl. intros v IN. cset_tac. apply types in H. rewrite lookup_set_spec in *.
destruct H as [x [INA1 HASVAL]]. exists x. split.
- cset_tac.
  + assumption.
  + apply H0. exists x. tauto.
- assumption.
Qed.

(* the leftover after a split from a panorama memory is well-typed *)
Lemma splitPanoramaRemainderTypes ET (ST:onv memory_type) G F E S U m A
  (ism:ST[m] = Some MemPanorama) (disj:alias_disjoint G A F) (tokens_apply: alias_tokens_apply E G):
  memoriesOfType ET ST G F E S U ->
  memoryOfType MemPanorama (F ∪ A) E (domain (restrict_mem (S[m]) (domain (S[m]) \ lookup_set E A))).
Proof.
intros types. invc types. rewrite restrict_mem_domain by crush. specialize (mems_type _ _ ism). simpl in *.
intros v [IN1 IN2]. cset_tac. clear H1. rewrite lookup_set_spec in *. destruct IN2 as [x [IN2 HASVAL]].
cset_tac. destruct IN2 as [IN2|IN2].
- apply (mems_type (E[x])). split.
  + assumption.
  + rewrite lookup_set_spec. exists x. tauto.
- apply H2. exists x. tauto.
Qed.


(* Having just a panorama memory types *)
Lemma memPanoramaOnlyTypes ET G E (S:env mem) U (m:var) (mv:mem) (inU:domain mv ⊆ U)
  (def: forall m, default (S[m]) = C_Unalloc) (hasval: S[m] = mv):
  memoriesOfType ET (ompty[m <- Some MemPanorama]) G BSet.empty E S U.
Proof.
apply MOT; unfold ompty in *; unfold Env.empty in *; intros.
- apply memPanoramaOnlyValid.
- eauto.
- crush.
- destruct_lookup_b. autoinjections; simpl. intros X [IN1 IN2]. rewrite lookup_set_spec in IN2.
  destruct IN2 as [x [INE _]]. crush.
Qed.


(** Well-typedness of programs *)
Inductive block_type := TB { (* Definition 10 (closure type = block type) *)
  TB_ET: onv ty; (* closure variables *)
  TB_ZT: list ty; (* parameter types *)
  TB_rt: ty (* return type *)
}.

Definition updET X (ET : onv X) (ZT : list X) (Z : params)
  : onv X := 
  update_with_list Z (List.map Some ZT) ET.

Inductive argsOfType (ET:onv ty) : list var -> list ty -> Prop :=
| WTA_nil : argsOfType ET nil nil
| WTA_cons x t Y YT : argsOfType ET Y YT -> ET [x] = Some t -> 
  argsOfType ET (x::Y) (t::YT).

(* Definition 11
   Note than ET is \Gamma, LT is \Lambda and ST is \Sigma  *)
(* The focus is an explicit member of the static context (which we call type_env here)
   to avoid having to compute it. This would require iterating over the list of elements
   of the domain of the memory environment, lookup up the types, extracting the
   set of variables and doing a union... it's easier to do this additional bookkeeping,
   and require it to stay consistent *)
Definition type_env:Type := onv ty * list block_type * onv memory_type * cset var.
Definition type_env_ET (te:type_env) := match te with (ET, _, _, _) => ET end.
Definition type_env_LT (te:type_env) := match te with (_, LT, _, _) => LT end.
Definition type_env_ST (te:type_env) := match te with (_, _, ST, _) => ST end.
Definition type_env_F (te:type_env) := match te with (_, _, _, F) => F end.

Definition empty_env (m:var): type_env := (* an empty type env, with m being the panorama memory *)
  (ompty, nil, (ompty[m <- Some MemPanorama]), BSet.empty).

Inductive stmtOfType : type_env -> stmt -> ty -> Prop := (* Figures 4.2 and 4.3 *)
| WT_Exp ET LT ST G F x e b t
    et (new:ET[x] = None)
    :  G ⊆ stmt_get_ann b
    -> expOfType ET e et -> ~isPtr et
    -> stmtOfType (ET[x<-Some et], LT, ST, F) b t
    -> stmtOfType (ET, LT, ST, F) (stmtExp G x e b) t

| WT_If ET LT ST G F x b1 b2 t
    et `(ET[x] = Some et)
    :  G ⊆ stmt_get_ann b1
    -> G ⊆ stmt_get_ann b2
    -> stmtOfType (ET, LT, ST, F) b1 t
    -> stmtOfType (ET, LT, ST, F) b2 t
    -> stmtOfType (ET, LT, ST, F) (stmtIf G x b1 b2) t

| WT_Goto ET LT ST G F l Y m t
    blk `(get LT (counted l) blk)
    (argT:argsOfType ET Y (TB_ZT blk))
    (memOK:ST[m] = Some MemPanorama)
    (focusEmpty: F = BSet.empty)
    (rtOk:TB_rt blk = t)
    : stmtOfType (ET, LT, ST, F) (stmtGoto G l Y m) t

| WT_Return ET LT ST G F x t
    `(ET[x] = Some t)
    : stmtOfType (ET, LT, ST, F) (stmtReturn G x) t

| WT_Let ET LT ST G F ET' s Z ZT m b t
    rt: G ⊆ stmt_get_ann b
    -> subEnv ET' ET
    -> stmtOfType (updET ET' ZT Z, (TB ET' ZT rt)::LT, ompty[m <- Some MemPanorama], BSet.empty) s rt
    -> stmtOfType (ET, (TB ET' ZT rt)::LT, ST, F) b t
    -> stmtOfType (ET, LT, ST, F) (stmtLet G Z m s b) t

| WT_Alloc ET LT ST G F a m b t
    `(ET[a] = None) `(ST[m] = None) G' `(G' = stmt_get_ann b) `(forall a', a' ∈ F -> alias G' (AliasNeq a a'))
    et: G ⊆ G'
    ->  stmtOfType (ET[a <- Some (ptrTo et)], LT, ST[m <- Some (MemFocus {{a}})], F ∪ {{a}}) b t
    ->  stmtOfType (ET, LT, ST, F) (stmtAlloc G a m b) t

| WT_Free ET LT ST G F m' m a b t
    `(ST[m <- None][m'] = None) mt `(ST[m] = Some mt) et `(ET[a] = Some (ptrTo et))
    :  in_mem G F mt a
    -> G ⊆ stmt_get_ann b
    -> stmtOfType (ET, LT, ST[m <- None][m' <- Some mt], F) b t
    -> stmtOfType (ET, LT, ST, F) (stmtFree G m' m a b) t

| WT_Store ET LT ST G F m' m a x b t
    `(ST[m <- None][m'] = None) mt `(ST[m] = Some mt) et `(ET[x] = Some et) `(ET[a] = Some (ptrTo et))
    :  in_mem G F mt a
    -> G ⊆ stmt_get_ann b
    -> stmtOfType (ET, LT, ST[m <- None][m' <- Some mt], F) b t
    -> stmtOfType (ET, LT, ST, F) (stmtStore G m' m a x b) t

| WT_Load ET LT ST G F x m a b t
    `(ET[x] = None) mt `(ST[m] = Some mt) et `(ET[a] = Some (ptrTo et))
    :  in_mem G F mt a
    -> G ⊆ stmt_get_ann b
    -> stmtOfType (ET[x <- Some et], LT, ST, F) b t
    -> stmtOfType (ET, LT, ST, F) (stmtLoad G x m a b) t

| WT_MergeF ET LT ST G F m'' m m' b t 
    `(m <> m') `(ST[m <- None][m' <- None][m''] = None)
    mA `(ST[m] = Some (MemFocus mA)) mA' `(ST[m'] = Some (MemFocus mA'))
    :  G = stmt_get_ann b
    -> stmtOfType (ET, LT, ST[m <- None][m' <- None][m'' <- Some (MemFocus (mA ∪ mA'))], F) b t
    -> stmtOfType (ET, LT, ST, F) (stmtMerge G m'' m m' b) t

| WT_MergeP ET LT ST G F m'' m m' b t 
    `(ST[m <- None][m' <- None][m''] = None)
    mA `(ST[m] = Some (MemFocus mA)) `(ST[m'] = Some MemPanorama)
    :  G = stmt_get_ann b
    -> stmtOfType (ET, LT, ST[m <- None][m' <- None][m'' <- Some MemPanorama], F\mA) b t
    -> stmtOfType (ET, LT, ST, F) (stmtMerge G m'' m m' b) t

| WT_SplitF ET LT ST G F m' m'' m A b t
    `(m' <> m'') `(ST[m <- None][m'] = None) `(ST[m <- None][m''] = None) mA `(ST[m] = Some (MemFocus mA))
    `(A ⊆ mA) `(alias_disjoint G A (mA\A))
    :  G = stmt_get_ann b
    -> stmtOfType (ET, LT, ST[m <- None][m' <- Some (MemFocus (mA\A))][m'' <- Some (MemFocus A)], F) b t
    -> stmtOfType (ET, LT, ST, F) (stmtSplit G m' m'' m A b) t

| WT_SplitP ET LT ST G F m' m'' m A b t
    `(m' <> m'') `(ST[m <- None][m'] = None) `(ST[m <- None][m''] = None) `(ST[m] = Some MemPanorama)
    `(forall a, a ∈ A -> exists t, ET[a] = Some (ptrTo t))
    `(alias_disjoint G A F)
    :  G = stmt_get_ann b
    -> stmtOfType (ET, LT, ST[m <- None][m' <- Some MemPanorama][m'' <- Some (MemFocus A)], F ∪ A) b t
    -> stmtOfType (ET, LT, ST, F) (stmtSplit G m' m'' m A b) t.


Definition labenvOfType (L:labenv) (LT: list block_type) : Prop := (* Definition 13 *)
    forall n ET ZT rt, get LT n (TB ET ZT rt) -> exists blk,
       get L n blk
    /\ stmtOfType (updET ET ZT (block_Z blk), drop n LT, ompty[block_m blk <- Some MemPanorama], BSet.empty) (block_s blk) rt
    /\ envOfType (block_E blk) ET.

Lemma labenvOfType_drop L LT n: labenvOfType L LT -> labenvOfType (drop n L) (drop n LT).
Proof.
revert L LT. induction n; simpl; intros.
- assumption.
- eapply IHn. clear IHn. destruct LT; simpl; [now isabsurd|]. unfold labenvOfType in *. intros ? ? ? ? GET1. apply getLS with (x':=b) in GET1. destruct (H _ _ _ _ GET1) as [blk [GET2 types]]. destruct L; simpl in *; [now isabsurd|]. exists blk. split.
  * invc GET2. assumption.
  * assumption.
Qed.

Lemma labenvOfType_upd L LT E ET Z ZT m s rt
  (Ed:envOfType E ET) 
  (sTD:stmtOfType (updET ET ZT Z, (TB ET ZT rt)::LT, ompty[m <- Some MemPanorama], BSet.empty) s rt)
  : labenvOfType L LT
  -> labenvOfType ((blockI E Z m s)::L) ((TB ET ZT rt)::LT).
Proof.
intros LOT. intros ? ? ? ? GET1.
destruct n; invc GET1; simpl in *.
- exists (blockI E Z m s). simpl in *. split.
  + apply getLB.
  + split; assumption.
- destruct (LOT _ _ _ _ H3) as [blk [GET2 types]]. exists blk. split.
  + apply getLS. assumption.
  + assumption.
Qed.


Lemma argsOfType_subEnv ET ET' Y Z
  : argsOfType ET' Y Z -> subEnv ET' ET -> argsOfType ET Y Z.
Proof.
  intros. general induction H; eauto using argsOfType.
Qed.

Lemma envOfType_listUpdate E ET E' ET' Z ZT Y: argsOfType ET Y ZT -> envOfType E ET -> envOfType E' ET' ->
  envOfType (updE E' E Z Y) (updET ET' ZT Z).
Proof.
revert ZT Y. unfold updE, updET. induction Z; simpl; intros ? ? argsT envE envE'.
- assumption.
- invc argsT; simpl.
  + assumption.
  + apply envOfType_update.
    * eapply IHZ; crush.
    * apply envE. assumption.
Qed.


(** The memory created by merging all typed memories *)
(* Basics: Merging a bunch of memories *)
Fixpoint merge_mems (l:list mem):mem := match l with
| nil => mempty
| mv::l => merge_mem mv (merge_mems l)
end.
Lemma merge_mems_mergehasval l v (c:cell): c <> C_Unalloc -> c = merge_mems l [v] ->
  exists mv:mem, In mv l /\ mv[v] = c.
Proof.
intros NONDEF. induction l; simpl; intros MERGEVAL.
- exfalso. crush.
- unfold merge_mem in *. destruct_prop (v ∈ (domain a)).
  + exists a. split.
    * left. reflexivity.
    * eapply mapjoin_spec in i. rewrite i in MERGEVAL. symmetry. assumption.
  + eapply mapjoin_spec in n. rewrite n in MERGEVAL. destruct (IHl MERGEVAL) as [mv [IN VAL]]. exists mv. tauto.
Qed.
Lemma merge_mems_memhasval l (c:cell):
  (forall mv1 mv2, In mv1 l -> In mv2 l -> mv1 <> mv2 -> disjoint (domain mv1) (domain mv2)) ->
  (forall mv, In mv l -> c <> default mv) ->
  forall mv v, In mv l -> mv[v] = c -> c = merge_mems l [v].
Proof.
induction l; simpl; intros DISJ NOTC ? ? IN HASVAL.
- isabsurd.
- unfold merge_mem. destruct IN as [EQ|IN].
  + subst a. rewrite <-HASVAL. symmetry. eapply mapjoin_spec. rewrite domain_spec2, HASVAL. eapply NOTC. crush.
  + assert (EQ:c = merge_mems l [v]). { eapply IHl; crush. }
    destruct_prop (a[v] = c).
    * assert (v ∈ domain a). { rewrite domain_spec2, e. eauto. }
      eapply mapjoin_spec in H. rewrite H. symmetry. assumption.
    * rewrite EQ. symmetry. apply mapjoin_spec. assert(v ∈ domain mv). { rewrite domain_spec2, HASVAL. eauto. }
      assert (mv <> a). { intros EQ'. subst a. isabsurd. }
      eapply disjoint_in_one with (A := domain mv); crush.
Qed.
Lemma merge_mems_indom (l:list mem):
  (forall mv, In mv l -> default mv = C_Unalloc) ->
  forall v, (C_Unalloc = merge_mems l [v] <-> forall mv, In mv l -> mv[v] = C_Unalloc).
Proof.
intros DEF v. split; revert DEF.
- induction l; simpl; intros.
  + isabsurd.
  + unfold merge_mem in *. assert (NOTIN:v ∉ domain a).
    { intros INDOM. assert (H1:=INDOM). eapply mapjoin_spec in H1. rewrite H1 in H. rewrite domain_spec2 in INDOM.
      apply INDOM. rewrite <-H. symmetry. eapply DEF; eauto. }
    destruct H0 as [EQ|IN].
   * subst a. rewrite domain_spec in NOTIN. rewrite NOTIN. eapply DEF; eauto.
   * eapply mapjoin_spec in NOTIN. rewrite NOTIN in H. eapply IHl; eauto.
- induction l; simpl; intros DEF ?.
  + crush.
  + unfold merge_mem. destruct_prop (v ∈ domain a).
    * eapply mapjoin_spec in i. rewrite i. symmetry. eapply H. eauto.
    * eapply mapjoin_spec in n. rewrite n. eapply IHl; eauto.
Qed.

(* A level of memory "well-typedness" which is sufficient for a noticeable amount of reasoning:
   If all memories are disjoint, then we can already argue how the merged-all memory behaves
   when certain operations are performed on individual memories  *)
Inductive memoriesDisjoint (ST:onv memory_type) (S:env mem):Prop :=
| MD
    (ST_default: default ST = None)
    (memories_disjoint: forall m1 t1 m2 t2, ST[m1] = Some t1 -> ST[m2] = Some t2 -> m1 <> m2 ->
                                            disjoint (domain (S[m1])) (domain (S[m2])))
    (memories_default: forall m, default (S[m]) = C_Unalloc)
    :memoriesDisjoint ST S.

Lemma welltypedMemories_Disjoint ET ST G F E S U:
  memoriesOfType ET ST G F E S U -> alias_tokens_apply E G -> memoriesDisjoint ST S.
Proof.
intros types apply. inv types. invc types_valid. split.
- assumption.
- intros ? ? ? ? IN1 IN2 INEQ. eapply welltypedMemoriesAreDisjoint; crush.
- assumption.
Qed.
Hint Resolve welltypedMemories_Disjoint.

Lemma memoriesDisjoint_forget ST S m: memoriesDisjoint ST S -> memoriesDisjoint (ST[m <- None]) S.
Proof.
intros disj. invc disj. split.
- crush.
- intros. crush.
- assumption.
Qed.

Lemma memoriesDisjoint_add (ST: onv memory_type) (S:env mem) m mt mv
  (newdisj: forall m1 mt1, ST[m1] = Some mt1 -> disjoint (domain (S[m1])) (domain mv))
  (def: default mv = C_Unalloc):
  memoriesDisjoint ST S -> memoriesDisjoint (ST[m <- Some mt]) (S[m <- mv]).
Proof.
intros disj. invc disj. split.
- crush.
- intros. crush.
- crush.
Qed.

(* Merging all memories having a type *)
Definition merge_all_mems (ST:onv memory_type) (S:env mem):mem :=
  merge_mems (elements (lookup_set S (domain ST))).

(* Propositional characterisation of merge_all_mems *)
Lemma merge_all_mems_hasval ST S v (c:cell):
  memoriesDisjoint ST S -> c <> C_Unalloc ->
  (c = merge_all_mems ST S [v] <-> exists m mt, ST[m] = Some mt /\ S[m][v] = c).
Proof.
intros disj INEQ. invc disj. split.
- intros MERGEVAL. unfold merge_all_mems in *. destruct (merge_mems_mergehasval _ INEQ MERGEVAL) as [mv [IN VAL]].
  rewrite <-elements_spec, lookup_set_spec in IN. destruct IN as [x [INDOM HASVAL]].
  exists x. rewrite domain_spec2, ST_default in INDOM. destruct (ST[x]).
  + exists m. crush.
  + exfalso. apply INDOM. reflexivity.
- intros [m [mt [HASTY HASVAL]]]. unfold merge_all_mems. apply merge_mems_memhasval with (mv := S[m]).
  + intros ? ? IN1 IN2 MINEQ. rewrite <-elements_spec, lookup_set_spec in IN1, IN2.
    destruct IN1 as [m1 [IN1 VAL1]], IN2 as [m2 [IN2 VAL2]]. subst mv1. subst mv2.
    eapply onv_domain in IN1; [|assumption]. eapply onv_domain in IN2; [|assumption]. destruct IN1, IN2.
    assert (m1 <> m2). { intros EQ. apply MINEQ. subst m1. reflexivity. }
    eapply memories_disjoint; eauto.
  + intros ? IN EQ. apply INEQ. rewrite EQ. rewrite <-elements_spec, lookup_set_spec in IN.
    destruct IN as [m2 [IN VAL]]. eapply onv_domain in IN; [|assumption]. destruct IN. subst mv.
    eapply memories_default.
  + rewrite <-elements_spec, lookup_set_spec. exists m. split.
    * rewrite domain_spec2. intros EQ. rewrite EQ, ST_default in HASTY. discriminate.
    * reflexivity.
  + assumption.
Qed.

Lemma merge_all_mems_hasnoval ST S v: memoriesDisjoint ST S ->
  (C_Unalloc = merge_all_mems ST S [v] <-> forall m mt, ST[m] = Some mt -> v ∉ domain (S[m])).
Proof.
intros disj. invc disj. unfold merge_all_mems. split.
- intros MERGEVAL ? ? HASTY. rewrite domain_spec, memories_default. eapply merge_mems_indom; try eassumption.
  + intros mv IN. rewrite <-elements_spec, lookup_set_spec in IN. destruct IN as [m' [IN VAL]].
    rewrite <-VAL. apply memories_default.
  + rewrite <-elements_spec, lookup_set_spec. exists m. split.
    * rewrite domain_spec2, HASTY, ST_default. discriminate.
    * reflexivity.
- intros NOTINDOM. eapply merge_mems_indom.
  + intros mv IN. rewrite <-elements_spec, lookup_set_spec in IN. destruct IN as [m' [IN VAL]].
    rewrite <-VAL. apply memories_default.
  + intros mv IN. rewrite <-elements_spec, lookup_set_spec in IN. destruct IN as [m' [IN VAL]].
    rewrite <-VAL, <-(memories_default m'), <-domain_spec. eapply onv_domain in IN; [|assumption].
    destruct IN. eapply NOTINDOM. eassumption.
Qed.

Lemma merge_all_mems_only_mem (m:var) mt (S:env mem) (def:forall m, default (S[m]) = C_Unalloc):
  merge_all_mems (ompty[m <- Some mt]) S === S[m].
Proof.
assert (disj: memoriesDisjoint (ompty[m <- Some mt]) S). { split; crush. }
intros x. symmetry. destruct_prop(S[m][x] = C_Unalloc).
- rewrite e. rewrite merge_all_mems_hasnoval by assumption. intros. destruct_lookup_b. rewrite domain_spec.
  rewrite def. assumption.
- rewrite merge_all_mems_hasval by assumption. exists m, mt. crush.
Qed.

(* hide the uglyness of using above lemmas when proving agreement between two merge_all_mems *)
Lemma merge_all_mems_agree (ST1 ST2: onv memory_type) (S1 S2: env mem) v
  (disj1: memoriesDisjoint ST1 S1) (disj2: memoriesDisjoint ST2 S2):
  (forall (NOTIN: forall m1 mt1, ST1[m1] = Some mt1 -> v ∉ domain (S1[m1])) m2 mt2,
     ST2[m2] = Some mt2 -> v ∉ domain (S2[m2])) ->
  (forall c m1 mt1, c <> C_Unalloc -> ST1[m1] = Some mt1 -> S1[m1][v] = c ->
                    exists m2 mt2, ST2[m2] = Some mt2 /\ S2[m2][v] = c) ->
  merge_all_mems ST1 S1 [v] = merge_all_mems ST2 S2 [v].
Proof.
intros UNALLOC_SAME ALLOC_SAME. remember (merge_all_mems ST1 S1 [v]) as c. destruct_prop(c=C_Unalloc).
- rewrite e in *. rewrite merge_all_mems_hasnoval by assumption. eapply UNALLOC_SAME.
  intros ? ? EQ. rewrite merge_all_mems_hasnoval in Heqc by assumption. eapply Heqc. eassumption.
- rewrite merge_all_mems_hasval by assumption. rewrite merge_all_mems_hasval in Heqc by assumption.
  destruct Heqc as [m [mt [EQ HASVAL]]]. eapply ALLOC_SAME; eassumption.
Qed.

Lemma merge_all_mems_equiv (ST1 ST2: onv memory_type) (S1 S2: env mem)
  (disj1: memoriesDisjoint ST1 S1) (disj2: memoriesDisjoint ST2 S2):
  (forall v (NOTIN: forall m1 mt1, ST1[m1] = Some mt1 -> v ∉ domain (S1[m1])) m2 mt2,
     ST2[m2] = Some mt2 -> v ∉ domain (S2[m2])) ->
  (forall v c m1 mt1, c <> C_Unalloc -> ST1[m1] = Some mt1 -> S1[m1][v] = c ->
                      exists m2 mt2, ST2[m2] = Some mt2 /\ S2[m2][v] = c) ->
  merge_all_mems ST1 S1 === merge_all_mems ST2 S2.
Proof.
intros UNALLOC_SAME ALLOC_SAME v.
eapply merge_all_mems_agree; eauto.
Qed.

(* Merging the same memories results in the same merged memory *)
Lemma merge_all_mems_same (ST1 ST2: onv memory_type) (S1 S2: env mem)
  (disj: memoriesDisjoint ST2 S2) (samedef: default ST1 = default ST2):
  ST1 === ST2 -> S1 === S2 -> merge_all_mems ST1 S1 === merge_all_mems ST2 S2.
Proof.
inv disj. intros EQ1 EQ2. apply merge_all_mems_equiv.
- split.
  + crush.
  + intros. rewrite EQ2. eapply memories_disjoint; [rewrite <-EQ1|rewrite <-EQ1|]; crush.
  + crush.
- assumption.
- intros. rewrite <-EQ2. eapply NOTIN. rewrite EQ1. eassumption.
- intros. exists m1, mt1. rewrite <-EQ1, <-EQ2. tauto.
Qed.

(* Merging two memories into one does not affect the merged-all memory *)
Lemma merge_all_mems_merge (ST: onv memory_type) S m1 mt1 m2 mt2 m3 mt3 (ineq: m1 <> m2)
  (ism1: ST[m1] = Some mt1) (ism2: ST[m2] = Some mt2) (new: ST[m1 <- None][m2 <- None][m3] = None)
  (disj: memoriesDisjoint ST S):
  merge_all_mems ST S === merge_all_mems (ST[m1 <- None][m2 <- None][m3 <- Some mt3])
                                         (S[m3 <- merge_mem (S[m1]) (S[m2])]).
Proof.
inv disj. apply merge_all_mems_equiv.
- assumption.
- apply memoriesDisjoint_add, memoriesDisjoint_forget, memoriesDisjoint_forget.
  + intros. rewrite merge_mem_domain by crush. apply disjoint_split; eapply memories_disjoint; crush.
  + rewrite merge_mem_default. eauto.
  + assumption.
- intros. destruct_prop (m0=m3); simplify lookup'.
  + rewrite merge_mem_domain by crush. intros IN. cset_tac. destruct IN as [IN|IN].
    * eapply (NOTIN m1); crush.
    * eapply (NOTIN m2); crush.
  + eapply NOTIN; crush.
- intros. destruct_prop(m0=m1 \/ m0=m2).
  + exists m3, mt3. simplify lookup'. split; [reflexivity|]. rewrite <-(memories_default m0), <-domain_spec2 in H.
    unfold merge_mem. destruct o; subst m0.
    * eapply mapjoin_spec in H. rewrite H. reflexivity.
    * assert (NIN:v ∉ domain (S[m1])). { intros IN. eapply memories_disjoint with (m1:=m1) (m2:=m2); eauto. }
      eapply mapjoin_spec in NIN. rewrite NIN. reflexivity.
  + exists m0, mt0. split; crush.
Qed.

(* Splitting a memory does not affect the merged-all memory *)
Lemma merge_all_mems_split (ST: onv memory_type) S m1 mt1 m2 mt2 m3 mt3 A
  (ism1: ST[m1] = Some mt1) (new2: ST[m1 <- None][m2] = None) (new3: ST[m1 <- None][m3] = None) (ineq: m2 <> m3)
  (disj: memoriesDisjoint ST S):
  merge_all_mems ST S === merge_all_mems (ST[m1 <- None][m2 <- Some mt2][m3 <- Some mt3])
                                         (S[m2 <- restrict_mem (S[m1]) (domain (S[m1]) \ A)]
                                           [m3 <- restrict_mem (S[m1]) A]).
Proof.
inv disj. apply merge_all_mems_equiv.
- assumption.
- apply memoriesDisjoint_add, memoriesDisjoint_add, memoriesDisjoint_forget.
  + intros. destruct_lookup.
    * (* the two restricted mems *) rewrite !restrict_mem_domain by crush.
      eapply disjoint_sym, disjoint_incl_lr, disjoint_minus with (A:=A); rewrite meet_comm;
      eapply meet_incl, incl_refl.
    * eapply disjoint_restrict; crush.
  + rewrite restrict_mem_default. reflexivity.
  + intros. eapply disjoint_restrict; crush.
  + rewrite restrict_mem_default. reflexivity.
  + assumption.
- intros. destruct_lookup; [|destruct_lookup_b].
  + rewrite restrict_mem_domain by crush. cset_tac. eapply NOTIN; crush.
  + rewrite restrict_mem_domain by crush. cset_tac. eapply NOTIN; crush.
- intros. destruct_prop(m0=m1).
  + destruct_prop (v ∈ A).
    * exists m3, mt3. simplify lookup'. split; [reflexivity|]. unfold restrict_mem. eapply update_list_upd.
      rewrite element_fromList. assumption.
    * exists m2, mt2. simplify lookup'. split; [reflexivity|]. unfold restrict_mem. eapply update_list_upd.
      rewrite element_fromList. rewrite <-(memories_default m1), <-domain_spec2 in H. cset_tac; tauto.
  + exists m0, mt0. crush.
Qed.

(* Renaming a memory and updating a location changes the merged-all memory only in that location *)
Lemma merge_all_mems_rename_update (ST: onv memory_type) (S: env mem) m mt m' v c
  (ism: ST[m] = Some mt) (new: ST[m <- None][m'] = None) (inm: S[m][v] <> C_Unalloc)
  (disj: memoriesDisjoint ST S):
  merge_all_mems ST S [v <- c] === merge_all_mems (ST[m <- None][m' <- Some mt]) (S[m' <- S[m][v <- c]]).
Proof.
inv disj. assert (disj':memoriesDisjoint (ST[m <- None][m' <- Some mt]) (S[m' <- S[m][v <- c]])).
{ apply memoriesDisjoint_add, memoriesDisjoint_forget.
  - intros. destruct_lookup_a. apply disjoint_incl_r with (C:=domain (S[m])).
    + eapply domain_update_incl. rewrite memories_default. assumption.
    + eapply memories_disjoint; crush.
  - crush.
  - assumption. }
intros x. destruct_prop(v = x); simplify lookup'.
- destruct_prop (c = C_Unalloc).
  + rewrite e in *. rewrite merge_all_mems_hasnoval by assumption. intros. destruct_lookup.
    * rewrite domain_spec, default_spec2. simplify lookup'. symmetry. eauto.
    * rewrite <-(memories_default m), <-domain_spec2 in inm. intros IN. eapply memories_disjoint; crush.
  + rewrite merge_all_mems_hasval by assumption. exists m', mt. crush.
- apply merge_all_mems_agree.
  + assumption.
  + assumption.
  + intros. destruct_lookup.
    * intros IN. eapply (NOTIN _ _ ism). eapply domain_update_incl; [|eassumption]. rewrite memories_default.
      assumption.
    * crush.
  + intros. destruct_prop(m1=m); simplify lookup'.
    * exists m', mt. crush.
    * exists m1, mt1. crush.
Qed.

(* If there is just one memory containing allocated cells, it's the same as the merged-all one *)
Lemma merge_all_mems_one_mem (ST: onv memory_type) (S:env mem) m1 mt1 m2 mt2
  (ispan: ST[m1] = Some mt1) (otherempty: forall m mt v, m <> m1 -> ST[m] = Some mt-> S[m][v] = C_Unalloc)
  (disj: memoriesDisjoint ST S):
  merge_all_mems ST S === merge_all_mems (ompty[m2 <- Some mt2]) (empty[m2 <- S[m1]]).
Proof.
inv disj. apply merge_all_mems_equiv.
- assumption.
- split; crush.
- intros. destruct_lookup.
  + eapply NOTIN. eassumption.
  + unfold empty, default_el, inst_defaulted_mem, mempty in *. intros IN.
    rewrite domain_spec2 in IN. simplify lookup'. rewrite default_spec in IN. apply IN. reflexivity.
- intros. exists m2, mt2. simplify lookup. split; [reflexivity|]. destruct_prop (m0=m1).
  + subst m0. reflexivity.
  + exfalso. apply H. eapply otherempty; eassumption.
Qed.

(* Adding a new singleton memory changs only hat location in the merged-all memory *)
Lemma merge_all_mems_new_mem (ST: onv memory_type) (S:env mem) m mt v
  (new: ST[m] = None) (newdisj: forall m mt, ST[m] = Some mt -> v ∉ domain(S[m]))
  (disj: memoriesDisjoint ST S):
  merge_all_mems ST S [v <- C_Uninit] === merge_all_mems (ST[m <- Some mt]) (S[m <- mempty[v <- C_Uninit]]).
Proof.
inv disj. assert (disj': memoriesDisjoint (ST[m <- Some mt]) (S[m <- mempty[v <- C_Uninit]])).
{ apply memoriesDisjoint_add.
  + intros. intros x [IN1 IN2]. rewrite domain_spec2 in IN2. crush.
  + crush.
  + assumption. }
intros x. destruct_prop(v = x); simplify lookup'.
- rewrite merge_all_mems_hasval by crush. exists m, mt. crush.
- apply merge_all_mems_agree.
  + assumption.
  + assumption.
  + intros. destruct_lookup.
    * rewrite domain_spec. simplify lookup'. rewrite default_spec2. crush.
    * eapply NOTIN; eauto.
  + intros. exists m1, mt1. split; crush.
Qed.

(* Chaging the type of a memory does not affect the merged-all memory *)
Lemma merge_all_mems_change_type ST S m mt0 mt1: default ST = None ->
  merge_all_mems (ST[m <- Some mt0]) S === merge_all_mems (ST[m <- Some mt1]) S.
Proof.
unfold merge_all_mems. intros ST_default.
assert (domain (ST[m <- Some mt0]) = domain (ST[m <- Some mt1])).
{ cset_tac; rewrite domain_spec2, default_spec2, ST_default in *; crush. }
rewrite H. reflexivity.
Qed.

(** Oracles *)
(* This is condition 6 of type-consistency (Definition 14) *)
Definition cellHasType (c:cell) (t:ty):Prop := c = C_Uninit \/ exists val, c = C val /\ valOfType val t.
Definition memory_oracle := map val (option ty).
Inductive oracleIsRight (ET:onv ty) (LT:list block_type) (s:memory_oracle) (E:env val) (L:labenv) (U:cset val) (mv:mem):Prop :=
  OIR
   (oracle_types:forall v t, s[v] = Some t -> v ∈ U /\ forall val, mv[v] = C val -> valOfType val t)
   (pointers_e_valid:forall a t, ET[a] = Some (ptrTo t) -> s[E[a]] = Some t)
   (pointers_lab_valid:forall a t blk tb n, get L n blk -> get LT n tb -> (TB_ET tb)[a] = Some (ptrTo t) -> s[(block_E blk)[a]] = Some t)
   (pointers_mem_valid:forall v t, s[v] = Some (ptrTo t) ->
                       forall val, mv[v] = C val -> s[val] = Some t)
   :oracleIsRight ET LT s E L U mv.
(* oracle_types without the U does not suffice: When proving oracleNewSingletonMem,
   we need to derive a contradiction from the oracle predicting a type for the newly allocated
   address. All we know is that this address is not in U, from which we can conclude that
   it is not in mv. *)

(* Some useful properties about valid oracles *)
Lemma oraclePtrInU (ET:onv ty) LT s (E:env val) L U mv
  (right: oracleIsRight ET LT s E L U mv) a ty (hasty: ET[a] = Some (ptrTo ty)):
  E[a] ∈ U.
Proof.
invc right. specialize (pointers_e_valid _ _ hasty).
specialize (oracle_types _ _ pointers_e_valid). tauto.
Qed.
  

Lemma oracleLoadTypes ET LT (ST:onv memory_type) G F (s:memory_oracle) E L (S:env mem) U
  (right:oracleIsRight ET LT s E L U (merge_all_mems ST S)) (types:memoriesOfType ET ST G F E S U)
  (tokens_apply: alias_tokens_apply E G)
  m a mt (mtypes:ST[m] = Some mt) et (valtypes:ET[a] = Some (ptrTo et)) v (memval:S[m][E[a]] = C v):
  valOfType v et.
Proof.
invc right. specialize (pointers_e_valid _ _ valtypes).
specialize (oracle_types _ _ pointers_e_valid).
assert (EQ: S[m][E[a]] = merge_all_mems ST S [E [a]]). { eapply merge_all_mems_hasval; crush. }
rewrite <-EQ in *. destruct oracle_types as [_ val_types]. apply val_types. assumption.
Qed.



(** State-type consistency: Definition 14 *)
Inductive stateConsistent: type_env -> memory_oracle -> state -> ty -> Prop :=
| SOT ET LT (ST:onv memory_type) F s E L S U b rt
   (stmt_types:stmtOfType (ET, LT, ST, F) b rt)
   (env_types:envOfType E ET)
   (labenv_types:labenvOfType L LT)
   (mems_type: memoriesOfType ET ST (stmt_get_ann b) F E S U)
   (oracle_right: oracleIsRight ET LT s E L U (merge_all_mems ST S))
   :stateConsistent (ET, LT, ST, F) s (E, L, S, U, b) rt.

(* Alias-correctness of a program: Definition 15 *)
Definition alias_tokens_correct (typeEnv:type_env) (b:stmt) (rt:ty):Prop
  := forall s E L S U, stateConsistent typeEnv s (E, L, S, U, b) rt ->
     forall st' E' G', fstepStar (E, L, S, U, b) st' -> E' = state_E st' -> G' = state_G st' ->
                       alias_tokens_apply E' G'.

(* The empty state is consistent with an empty static context *)
Lemma empty_state_of_empty_type (m:var) (b:stmt) (rt:ty): stmtOfType (empty_env m) b rt
  -> stateConsistent (empty_env m) (@BMap.empty val (option ty) None) (empty_state b) rt.
Proof.
intros stmt_types. apply SOT; unfold ompty in *; unfold Env.empty in *; unfold mempty in *.
- now apply stmt_types.
- now apply envOfType_empty.
- intros n ET ZT rt' GET. now inversion GET.
- eapply memPanoramaOnlyTypes with (mv := mempty).
  + intros x IN. exfalso. apply domain_spec2 in IN. crush.
  + crush.
  + crush.
- apply OIR; crush.
Qed.

(** Synthesis of consistent states *)
(* Given a well-typed program and a valid memory environment, we can synthesise a consistent state.
   For now, the program must not have free functions (i.e. the label type environment must be empty)  *)

(* generate values and oracle for the given types *)
Fixpoint gen_values_and_oracle (ET: onv ty) (X: list var): env val * memory_oracle :=
  match X with
  | nil => (empty, @BMap.empty val (option ty) None)
  | x::X => let (E, s) := gen_values_and_oracle ET X in
            (* extend E and s by definitions for x, as necessary *)
            match (ET[x]) with
            | None => (E, s) (* x does not have a type, ignore *)
            | Some t => (* check x's type: pointer? *)
              (match isPtr_sigma t with
               | inleft (exist pt _) => (* it's a pointer, generate address and extend oracle *)
                 let a := allocator (domain s) in
                 (E[x <- a], s[a <- Some pt])
               | inright _ => (* it's not a pointer, synthesise value *)
                 (E[x <- proj1_sig (valSynth t)], s) end)
            end
  end.

Lemma gen_default ET X: default (snd (gen_values_and_oracle ET X)) = None.
Proof.
induction X; simpl.
- crush.
- destruct (gen_values_and_oracle ET X) as [E s]. destruct (ET[a]).
  + destruct (isPtr_sigma t) as [[pt EQ]|NEQ].
    * simpl. crush.
    * simpl. assumption.
  + assumption.
Qed.

Lemma gen_hasty ET X: unique X -> forall x t, In x X -> ET[x] = Some t ->
  valOfType ((fst (gen_values_and_oracle ET X)[x])) t.
Proof.
induction X; intros UN x t IN HASTY; simpl in *.
- contradiction IN.
- destruct UN as [NOTIN UN]. destruct IN as [EQ|IN].
  + destruct (gen_values_and_oracle ET X) as [E s]. subst a. destruct (ET[x]); [|discriminate].
    destruct (isPtr_sigma t0) as [[pt EQ]|NEQ]; simpl; simplify lookup'.
    * autoinjections. eapply allocatorGivesPtr.
    * destruct (valSynth t0) as [v VHASTY]. simpl. autoinjections. assumption.
  + specialize (IHX UN _ _ IN HASTY). unfold fresh in *.
    destruct (gen_values_and_oracle ET X) as [E s]. simpl in *. destruct (ET[a]);[|crush].
    assert (x <> a). { intros EQ. subst. eapply NOTIN. assumption. }
    destruct (isPtr_sigma t0) as [[pt EQ]|NEQ]; crush.
Qed.

Lemma gen_oracle ET X: unique X -> forall x t, In x X -> ET[x] = Some (ptrTo t) -> 
  match gen_values_and_oracle ET X with (E, s) => s[E[x]] = Some t end.
Proof.
induction X; intros UN x t IN HASTY; simpl in *.
- crush.
- destruct UN as [NOTIN UN]. destruct IN as [EQ|IN].
  + destruct (gen_values_and_oracle ET X) as [E s]. subst a. destruct (ET[x]); [|discriminate].
    destruct (isPtr_sigma t0) as [[pt EQ]|NEQ]; simpl; simplify lookup'.
    * autoinjections. f_equal. apply ptrTo_injective. assumption.
    * exfalso. eapply NEQ. autoinjections. reflexivity.
  + specialize (IHX UN _ _ IN HASTY). unfold fresh in *. assert (DEF:= gen_default ET X).
    destruct (gen_values_and_oracle ET X) as [E s]. destruct (ET[a]);[|crush].
    assert (x <> a). { intros EQ. subst. eapply NOTIN. assumption. }
    destruct (isPtr_sigma t0) as [[pt EQ]|NEQ]; simpl in *; simplify lookup'.
    * assert (E[x] <> allocator (domain s)).
      { intros EQ. apply (allocatorGivesNew (domain s)). rewrite <-EQ. rewrite domain_spec2.
        rewrite DEF, IHX. discriminate. }
      simplify lookup'. assumption.
    * assumption.
Qed.

Lemma synth_state (ET: onv ty) (ET_def: default ET = None) (ST: onv memory_type) F b t
  (ST_valid: memTypesValid ET ST (stmt_get_ann b) F) (prg_types: stmtOfType (ET, nil, ST, F) b t):
  exists s E L S U, stateConsistent (ET, nil, ST, F) s (E, L, S, U, b) t.
Proof.
assert (s_default := gen_default ET (elements (domain ET))).
assert (ET_hasty := @gen_hasty ET (elements (domain ET))).
assert (s_correct := @gen_oracle ET (elements (domain ET))).
destruct (gen_values_and_oracle ET (elements (domain ET))) as [E s].
exists s, E, nil, empty, (domain s). unfold empty, default_el. simpl in *. unfold mempty.
constructor.
- assumption.
- intros x t' hasty. apply ET_hasty.
  + apply elements_unique.
  + apply elements_spec. rewrite domain_spec2, ET_def, hasty. discriminate.
  + assumption.
- crush.
- constructor.
  + assumption.
  + crush.
  + intros. simplify lookup'. intros x IN. exfalso. rewrite domain_spec2 in IN. apply IN. crush.
  + intros. destruct mt; simpl; simplify lookup'.
    * intros x IN. exfalso. rewrite domain_spec2 in IN. apply IN. crush.
    * intros x IN. exfalso. rewrite domain_spec2 in IN. apply IN. crush.
- assert (DISJ:memoriesDisjoint ST (∅ (∅ C_Unalloc))).
  { invc ST_valid. constructor.
    - crush.
    - intros. intros x [A _]. rewrite domain_spec2 in A. apply A. crush.
    - crush. }
  constructor.
  + intros ? ? hasty. split.
    * rewrite domain_spec2, hasty, s_default. discriminate.
    * intros ? ISVAL. exfalso. symmetry in ISVAL. eapply merge_all_mems_hasval in ISVAL; crush.
  + intros a t' hasty. apply s_correct.
    * apply elements_unique.
    * apply elements_spec. rewrite domain_spec2, ET_def, hasty. discriminate.
    * assumption.
  + intros. inversion H.
  + intros ? ? hasval ? ISVAL. exfalso. symmetry in ISVAL. eapply merge_all_mems_hasval in ISVAL; crush.
Qed.



(* 
*** Local Variables: ***
*** coq-load-path: ("infra" "constr" "il" "isa" ".") ***
*** End: ***
*)
